# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import glob
import os
import random
import shutil
import unittest

from availsim4core import main
from availsim4core.test.E2E.E2E_utils import Result


class test_E2E_failure_on_demand(unittest.TestCase):
    """
    Failure mode can be defined using mttf following different distribution functions or can be defined as "on demand"
    """

    ###
    expected_result_scenario_1 = [
        Result(11, "A_0_2", "FIRST_PHASE", "FAILED", "_MEAN_OCCURRENCES"),
        Result(11, "A_0_2", "SECOND_PHASE", "FAILED", "_MEAN_OCCURRENCES"),
        Result(11, "A_0_2", "SECOND_PHASE", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
    ]
    for component_name in ["Phase", "ROOT_0_1", "A_0_2"]:
        expected_result_scenario_1.append(Result(100, component_name, "*", "*", "_MEAN_DURATION", tolerance=6))
    scenario_1 = \
        ["./availsim4core/test/E2E/input/failure_on_demand/simulation.xlsx",
         "./availsim4core/test/E2E/input/failure_on_demand/pofod_first_phase__system.xlsx",
         None,
         1,
         expected_result_scenario_1]

    ###
    expected_result_scenario_2 = [
        Result(10, "A_0_2", "SECOND_PHASE", "FAILED", "_MEAN_OCCURRENCES"),
        Result(10, "A_0_2", "FIRST_PHASE", "FAILED", "_MEAN_OCCURRENCES"),
        Result(10, "A_0_2", "FIRST_PHASE", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
    ]
    for component_name in ["Phase", "ROOT_0_1", "A_0_2"]:
        expected_result_scenario_2.append(Result(100, component_name, "*", "*", "_MEAN_DURATION", tolerance=6))
    scenario_2 = \
        ["./availsim4core/test/E2E/input/failure_on_demand/simulation.xlsx",
         "./availsim4core/test/E2E/input/failure_on_demand/pofod_second_phase__system.xlsx",
         None,
         1,
         expected_result_scenario_2]

    ###
    expected_result_scenario_3 = [
        Result(0, "A_0_2", "FIRST_PHASE", "FAILED", "_MEAN_OCCURRENCES"),
        Result(0, "A_0_2", "SECOND_PHASE", "FAILED", "_MEAN_OCCURRENCES"),
        Result(0, "A_0_2", "SECOND_PHASE", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
    ]
    for component_name in ["Phase", "ROOT_0_1", "A_0_2"]:
        expected_result_scenario_3.append(Result(100, component_name, "*", "*", "_MEAN_DURATION", tolerance=6))
    scenario_3 = \
        ["./availsim4core/test/E2E/input/failure_on_demand/simulation.xlsx",
         "./availsim4core/test/E2E/input/failure_on_demand/pofod_no_phase__system.xlsx",
         None,
         1,
         expected_result_scenario_3]

    ###
    expected_result_scenario_4 = [
        Result(0, "A_0_2", "FIRST_PHASE", "FAILED", "_MEAN_OCCURRENCES"),
        Result(0, "A_0_2", "SECOND_PHASE", "FAILED", "_MEAN_OCCURRENCES"),
        Result(0, "A_0_2", "SECOND_PHASE", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
    ]
    for component_name in ["Phase", "ROOT_0_1", "A_0_2"]:
        expected_result_scenario_4.append(Result(100, component_name, "*", "*", "_MEAN_DURATION", tolerance=6))
    scenario_4 = \
        ["./availsim4core/test/E2E/input/failure_on_demand/simulation.xlsx",
         "./availsim4core/test/E2E/input/failure_on_demand/pofod_first_phase_no_failure__system.xlsx",
         None,
         1,
         expected_result_scenario_4]

    ###
    expected_result_scenario_5 = [
        Result(0, "A_0_2", "SECOND_PHASE", "FAILED", "_MEAN_OCCURRENCES"),
        Result(0, "A_0_2", "FIRST_PHASE", "FAILED", "_MEAN_OCCURRENCES"),
        Result(0, "A_0_2", "FIRST_PHASE", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
    ]
    for component_name in ["Phase", "ROOT_0_1", "A_0_2"]:
        expected_result_scenario_5.append(Result(100, component_name, "*", "*", "_MEAN_DURATION", tolerance=6))
    scenario_5 = \
        ["./availsim4core/test/E2E/input/failure_on_demand/simulation.xlsx",
         "./availsim4core/test/E2E/input/failure_on_demand/pofod_second_phase_no_failure__system.xlsx",
         None,
         1,
         expected_result_scenario_5]

    ###
    expected_result_scenario_6 = [
        Result(0, "A_0_2", "FIRST_PHASE", "FAILED", "_MEAN_OCCURRENCES"),
        Result(0, "A_0_2", "SECOND_PHASE", "FAILED", "_MEAN_OCCURRENCES"),
        Result(0, "A_0_2", "SECOND_PHASE", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
    ]
    for component_name in ["Phase", "ROOT_0_1", "A_0_2"]:
        expected_result_scenario_6.append(Result(100, component_name, "*", "*", "_MEAN_DURATION", tolerance=6))
    scenario_6 = \
        ["./availsim4core/test/E2E/input/failure_on_demand/simulation.xlsx",
         "./availsim4core/test/E2E/input/failure_on_demand/pofod_no_phase_no_failure__system.xlsx",
         None,
         1,
         expected_result_scenario_6]

    ###
    expected_result_scenario_7 = [
        Result(349, "A_0_2", "FIRST_PHASE", "FAILED", "_MEAN_OCCURRENCES"),
        Result(349, "A_0_2", "SECOND_PHASE", "FAILED", "_MEAN_OCCURRENCES"),
        Result(349, "A_0_2", "SECOND_PHASE", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
    ]
    for component_name in ["Phase", "ROOT_0_1", "A_0_2"]:
        expected_result_scenario_7.append(Result(100, component_name, "*", "*", "_MEAN_DURATION", tolerance=6))
    scenario_7 = \
        ["./availsim4core/test/E2E/input/failure_on_demand/simulation.xlsx",
         "./availsim4core/test/E2E/input/failure_on_demand/pofod_first_phase_some_failures__system.xlsx",
         None,
         1,
         expected_result_scenario_7]

    ###
    expected_result_scenario_8 = [
        Result(348, "A_0_2", "SECOND_PHASE", "FAILED", "_MEAN_OCCURRENCES"),
        Result(348, "A_0_2", "FIRST_PHASE", "FAILED", "_MEAN_OCCURRENCES"),
        Result(348, "A_0_2", "FIRST_PHASE", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
    ]
    for component_name in ["Phase", "ROOT_0_1", "A_0_2"]:
        expected_result_scenario_8.append(Result(100, component_name, "*", "*", "_MEAN_DURATION", tolerance=6))
    scenario_8 = \
        ["./availsim4core/test/E2E/input/failure_on_demand/simulation.xlsx",
         "./availsim4core/test/E2E/input/failure_on_demand/pofod_second_phase_some_failures__system.xlsx",
         None,
         1,
         expected_result_scenario_8]

    ###
    expected_result_scenario_9 = [
        Result(0, "A_0_2", "FIRST_PHASE", "FAILED", "_MEAN_OCCURRENCES"),
        Result(0, "A_0_2", "SECOND_PHASE", "FAILED", "_MEAN_OCCURRENCES"),
        Result(0, "A_0_2", "SECOND_PHASE", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
    ]
    for component_name in ["Phase", "ROOT_0_1", "A_0_2"]:
        expected_result_scenario_9.append(Result(100, component_name, "*", "*", "_MEAN_DURATION", tolerance=6))
    scenario_9 = \
        ["./availsim4core/test/E2E/input/failure_on_demand/simulation.xlsx",
         "./availsim4core/test/E2E/input/failure_on_demand/pofod_no_phase_some_failures__system.xlsx",
         None,
         1,
         expected_result_scenario_9]

    ###
    expected_result_scenario_10 = [
        Result(125, "MAIN_0_2", "OPERATION", "FAILED", "_MEAN_OCCURRENCES"),
        Result(125, "MAIN_0_2", "DEGRADED", "FAILED", "_MEAN_OCCURRENCES"),
        Result(125, "MAIN_0_2", "DEGRADED", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
        Result(62, "MAIN_0_2", "OPERATION", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
        Result(63, "MAIN_0_2", "SHUTDOWN", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
        Result(125, "KICK_0_3", "DEGRADED", "RUNNING", "_MEAN_OCCURRENCES"),
        Result(125, "KICK_0_3", "DEGRADED", "RUNNING", "_MEAN_OCCURRENCES"),
        Result(63, "KICK_0_3", "DEGRADED", "FAILED", "_MEAN_OCCURRENCES"),
        Result(63, "KICK_0_3", "SHUTDOWN", "FAILED", "_MEAN_OCCURRENCES"),
        Result(63, "KICK_0_3", "SHUTDOWN", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
    ]
    for component_name in ["Phase", "ROOT_0_1", "MAIN_0_2", "KICK_0_3"]:
        expected_result_scenario_10.append(Result(100, component_name, "*", "*", "_MEAN_DURATION", tolerance=6))
    scenario_10 = \
        ["./availsim4core/test/E2E/input/failure_on_demand/simulation.xlsx",
         "./availsim4core/test/E2E/input/failure_on_demand/pofod_second_basic_in_other_phase__system.xlsx",
         None,
         1,
         expected_result_scenario_10]


    param_scenario = [
        scenario_1,
        scenario_2,
        scenario_3,
        scenario_4,
        scenario_5,
        scenario_6,
        scenario_7,
        scenario_8,
        scenario_9,
        scenario_10,
    ]

    def test_runner(self):
        # randomize the order of the test execution to detect hidden dependencies
        random.shuffle(self.param_scenario)
        for simulation_file, system_file, sensitivity_file, expected_number_file_result, expected_result_list in self.param_scenario:
            with self.subTest():
                self._runner_E2E(simulation_file,
                                 system_file,
                                 sensitivity_file,
                                 expected_number_file_result,
                                 expected_result_list)

    def _runner_E2E(self,
                    simulation_file,
                    system_file,
                    sensitivity_file,
                    expected_number_file_result,
                    expected_result_list):

        output_folder = "./availsim4core/test/E2E/output/failure_on_demand/"

        # Clean folder
        if os.path.exists(output_folder):
            shutil.rmtree(output_folder)

        # Run the main process
        main.start(simulation_file, system_file, sensitivity_file, output_folder)

        result_simulation_file_list = glob.glob(output_folder + "/*.xlsx")
        self.assertEqual(len(result_simulation_file_list), expected_number_file_result)

        for expected_result in expected_result_list:
            actual_result = expected_result.extract_result(result_simulation_file_list)

            if expected_result.tolerance == -1:
                self.assertEqual(expected_result.expected_result, actual_result)
            else:
                self.assertAlmostEqual(expected_result.expected_result, actual_result, places=expected_result.tolerance)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from __future__ import annotations

from typing import List

from availsim4core.src.context.context import Context
from availsim4core.src.context.system.children_logic.oo import Oo
from availsim4core.src.context.system.component_tree.component import Component
from availsim4core.src.context.system.component_tree.status import Status


class RequiredComponent(Oo):

    def __init__(self, minimum_number_of_required_component: int):
        super().__init__(minimum_number_of_required_component)

    def __repr__(self):
        return f"{self.minimum_number_of_required_component}RC"

    def evaluate(self, list_of_children: List[Component], context: Context) -> Status:
        return super().evaluate(list_of_children, context)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from __future__ import annotations

from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from availsim4core.src.context.system.component_tree.basic import Basic

from availsim4core.src.context.context import Context
from availsim4core.src.context.system.failure_mode import FailureMode
from availsim4core.src.discrete_event_simulation.event.b_event.basic_b_event import BasicBEvent


class FailureEvent(BasicBEvent):
    """
    Defines a failure event associated to a Component and a Failure Mode.
    """
    __slots__ = 'failure_mode'

    def __init__(self,
                 absolute_occurrence_time: float,
                 context: Context,
                 basic: Basic,
                 failure_mode: FailureMode,
                 priority):
        super().__init__(absolute_occurrence_time, context, basic, priority)
        self.failure_mode = failure_mode
        self.failure_mode.sample_generation_time = context.absolute_simulation_time

    def __hash__(self):
        return hash((type(self), self.absolute_occurrence_time, self.priority, self.basic, self.failure_mode))

    def __eq__(self, other):
        # TODO: when only one failure_mode is present in a basic, the comparison should not be on the failure mode anymore
        if not isinstance(other, type(self)):
            return NotImplemented
        return super().__eq__(other) and \
               self.failure_mode == other.failure_mode

    def __str__(self):
        return (f"type ::{type(self)} "
                f"basic ::{self.basic.name}, "
                f"lid ::{self.basic.local_id}, "
                f"gid ::{self.basic.global_id}, "
                f"at t ::{self.absolute_occurrence_time}, "
                f"FailureMode ::{self.failure_mode}\n")

    def is_context_valid(self):
        return self.context.phase_manager.current_phase in self.failure_mode.phase_set

    @staticmethod
    def _b_events_to_be_cleaned() -> List:
        from availsim4core.src.discrete_event_simulation.event.b_event.phase_event.next_phase_event import \
            NextPhaseEvent
        from availsim4core.src.discrete_event_simulation.event.b_event.phase_event.next_phase_if_failure_event import \
            NextPhaseIfFailureEvent
        from availsim4core.src.discrete_event_simulation.event.b_event.phase_event.jump_phase_event import \
            JumpPhaseEvent
        return [NextPhaseEvent, NextPhaseIfFailureEvent, JumpPhaseEvent]



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest

from availsim4core.src.context.context import Context
from availsim4core.src.context.system.children_logic.and_ import And
from availsim4core.src.context.system.component_tree.compound import Compound
from availsim4core.src.context.system.failure import Failure, FailureType
from availsim4core.src.context.system.failure_mode import FailureMode
from availsim4core.src.context.system.probability_law.deterministic_law import DeterministicLaw
from availsim4core.src.discrete_event_simulation.event.b_event.failure_event.blind_failure_event import \
    BlindFailureEvent
from availsim4core.src.discrete_event_simulation.event.b_event.failure_event.detectable_failure_event import \
    DetectableFailureEvent
from availsim4core.src.discrete_event_simulation.event.b_event.failure_event.failure_event_factory import \
    FailureEventFactory


class test_FailureEventFactory(unittest.TestCase):

    def test_build_detectable(self):
        DETERMINISTIC_VALUE = 1
        context = Context(None, [], set())
        failure_mode_1 = FailureMode("test_run",
                                     DeterministicLaw([DETERMINISTIC_VALUE]),
                                     DeterministicLaw([DETERMINISTIC_VALUE]),
                                     Failure(FailureType.DETECTABLE),
                                     [],
                                     None,
                                     [],
                                     None,
                                     'NEVER',
                                     [])

        root_component = Compound(1, "ROOT", 0, [], And(), [])

        ABSOLUTE_SIMULATION_TIME = 5
        result = FailureEventFactory.build(DETERMINISTIC_VALUE + ABSOLUTE_SIMULATION_TIME,
                                           context, root_component, failure_mode_1)

        expected_result = DetectableFailureEvent(DETERMINISTIC_VALUE + ABSOLUTE_SIMULATION_TIME,
                                                 context,
                                                 root_component,
                                                 failure_mode_1)

        self.assertEqual(result, expected_result)

    def test_build_blind(self):
        DETERMINISTIC_VALUE = 1
        context = Context(None, [], set())
        failure_mode_1 = FailureMode("test_run",
                                     DeterministicLaw([DETERMINISTIC_VALUE]),
                                     DeterministicLaw([DETERMINISTIC_VALUE]),
                                     Failure(FailureType.BLIND),
                                     [],
                                     None,
                                     [],
                                     None,
                                     'NEVER',
                                     [])

        root_component = Compound(1, "ROOT", 0, [], And(), [])

        ABSOLUTE_SIMULATION_TIME = 5
        result = FailureEventFactory.build(DETERMINISTIC_VALUE + ABSOLUTE_SIMULATION_TIME,
                                           context, root_component, failure_mode_1)

        expected_result = BlindFailureEvent(DETERMINISTIC_VALUE + ABSOLUTE_SIMULATION_TIME,
                                            context,
                                            root_component,
                                            failure_mode_1)

        self.assertEqual(result, expected_result)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from typing import List, Set

import pandas

from availsim4core.src.analysis import Analysis
from availsim4core.src.context.phase.phase import Phase
from availsim4core.src.context.system.failure_mode import FailureMode
from availsim4core.src.context.system.failure_mode_assignments import FailureModeAssignments
from availsim4core.src.context.system.inspection import Inspection
from availsim4core.src.context.system.minimal_replaceable_unit import MinimalReplaceableUnit
from availsim4core.src.context.system.phase_jump_trigger import PhaseJumpTrigger
from availsim4core.src.context.system.rca_trigger import RootCauseAnalysisTrigger
from availsim4core.src.context.system.system_template import SystemTemplate
from availsim4core.src.reader.xlsx.monte_carlo_reader import SimulationMonteCarloColumn
from availsim4core.src.reader.xlsx.simulation_reader import SimulationSheet
from availsim4core.src.reader.xlsx.system_template_reader import SystemTemplateInspectionsColumn, \
    SystemTemplateMinimalReplaceableUnitColumn, SystemTemplateFailureModeAssignmentsColumn, \
    SystemTemplateFailureModeColumn, SystemTemplateArchitectureColumn, SystemTemplateSheet, SystemTemplateField, \
    SystemTemplatePhasesColumn, SystemTemplateRootCauseAnalysisColumn, SystemTemplatePhaseJumpColumn
from availsim4core.src.simulation.monte_carlo import MonteCarlo
from availsim4core.src.simulation.simulation import Simulation


class XLSXAnalysisExporter:

    @classmethod
    def export(cls,
               writer,
               analysis:Analysis):
        cls.export_system(writer,
                          analysis.system_template)
        cls.export_simulation(writer,
                              analysis.simulation)

    @classmethod
    def export_system(cls,
                      writer,
                      system_template:SystemTemplate):

        dataframe = cls._export_architecture_entry_to_dataframe(
            system_template.architecture_entry_list)
        dataframe.to_excel(writer, sheet_name=SystemTemplateSheet.ARCHITECTURE, index=False)

        dataframe = cls._export_failure_mode_to_dataframe(
            system_template.failure_mode_list)
        dataframe.to_excel(writer, sheet_name=SystemTemplateSheet.FAILURE_MODES, index=False)

        dataframe = cls._export_failure_mode_assignments_to_dataframe(
            system_template.failure_mode_assignments_list)
        dataframe.to_excel(writer, sheet_name=SystemTemplateSheet.FAILURE_MODE_ASSIGNMENTS, index=False)

        dataframe = cls._export_mru_to_dataframe(
            system_template.mru_list)
        dataframe.to_excel(writer, sheet_name=SystemTemplateSheet.MINIMAL_REPLACEABLE_UNIT, index=False)

        dataframe = cls._export_inspection_to_dataframe(
            system_template.inspection_list)
        dataframe.to_excel(writer, sheet_name=SystemTemplateSheet.INSPECTIONS, index=False)

        dataframe = cls._export_phase_to_dataframe(
            system_template.phase_set)
        dataframe.to_excel(writer, sheet_name=SystemTemplateSheet.PHASES, index=False)

        dataframe = cls._export_rca_to_dataframe(
            system_template.root_cause_analysis_trigger_set)
        dataframe.to_excel(writer, sheet_name=SystemTemplateSheet.ROOT_CAUSE_ANALYSIS, index=False)

        dataframe = cls._export_phase_jump_to_dataframe(
            system_template.phase_jump_trigger_set)
        dataframe.to_excel(writer, sheet_name=SystemTemplateSheet.PHASE_JUMP, index=False)

    @classmethod
    def export_simulation(cls,
                          writer,
                          simulation:Simulation):
        dataframe = cls._export_simulation_to_dataframe(simulation)
        dataframe.to_excel(writer, sheet_name=SimulationSheet.SHEET, index=False)

    @classmethod
    def _export_architecture_entry_to_dataframe(cls, architecture_entry_list):
        data = []
        for architecture_entry in architecture_entry_list:
            data_architecture = [architecture_entry.component_name,
                                 architecture_entry.component_type_str,
                                 architecture_entry.component_number,
                                 architecture_entry.children_name_list if architecture_entry.children_name_list
                                 else SystemTemplateField.NONE,
                                 architecture_entry.children_logic_str if architecture_entry.children_logic_str
                                 else SystemTemplateField.NONE,
                                 architecture_entry.in_mru_str_list if architecture_entry.in_mru_str_list
                                 else SystemTemplateField.NONE,
                                 architecture_entry.trigger_mru_str_list if architecture_entry.trigger_mru_str_list
                                 else SystemTemplateField.NONE]
            data.append(data_architecture)

        column_names = cls.extract_value_list_from_class_public_attributes(SystemTemplateArchitectureColumn)
        dataframe = pandas.DataFrame(data, columns=column_names)
        return dataframe

    @staticmethod
    def extract_value_list_from_class_public_attributes(class_name):
        """
        Utility method to extract values from public attributes for a given class.
        Useful for constant global classes.
        """
        return [getattr(class_name, name)
                for name in vars(class_name).keys()
                if not name.startswith("__")]

    @classmethod
    def _export_failure_mode_to_dataframe(cls, failure_mode_list:List[FailureMode]):
        data = []
        for failure_mode in failure_mode_list:
            data_failure = [
                failure_mode.name,
                failure_mode.failure_law.name,
                failure_mode.failure_law.parameters,
                failure_mode.repair_law.name,
                failure_mode.repair_law.parameters,
                failure_mode.failure.type_of_failure,
                str(sorted([phase.name for phase in failure_mode.held_before_repair_phase_set])),
                SystemTemplateField.NONE if not failure_mode.inspection else failure_mode.inspection.name,
                str(sorted([phase.name for phase in failure_mode.phase_set])),
                SystemTemplateField.NONE if failure_mode.failure_mode_next_phase_if_failure is None else failure_mode.failure_mode_next_phase_if_failure.name,
                failure_mode.phase_change_trigger,
                str(sorted([phase.name for phase in failure_mode.held_after_repair_phase_set]))
            ]
            data.append(data_failure)

        column_names = cls.extract_value_list_from_class_public_attributes(SystemTemplateFailureModeColumn)
        dataframe = pandas.DataFrame(data, columns=column_names)
        return dataframe

    @classmethod
    def _export_failure_mode_assignments_to_dataframe(cls, failure_mode_assignments_list:List[FailureModeAssignments]):
        data = []
        for failure_mode_assignments in failure_mode_assignments_list:
            data_failure_assignments = [failure_mode_assignments.component_name,
                                        failure_mode_assignments.failure_mode.name]
            data.append(data_failure_assignments)

        column_names = cls.extract_value_list_from_class_public_attributes(SystemTemplateFailureModeAssignmentsColumn)
        dataframe = pandas.DataFrame(data, columns=column_names)
        return dataframe

    @classmethod
    def _export_mru_to_dataframe(cls, mru_list:List[MinimalReplaceableUnit]):
        data = []
        for mru in mru_list:
            data_mru = [mru.name,
                        mru.repair_law.name,
                        mru.repair_law.parameters,
                        mru.repair_schedule,
                        mru.scope_common_ancestor,
                        mru.status]
            data.append(data_mru)

        column_names = cls.extract_value_list_from_class_public_attributes(SystemTemplateMinimalReplaceableUnitColumn)
        dataframe = pandas.DataFrame(data, columns=column_names)
        return dataframe

    @classmethod
    def _export_inspection_to_dataframe(cls, inspection_list:List[Inspection]):
        data = []
        for inspection in inspection_list:
            data_inspection = [inspection.name,
                               inspection.periodicity,
                               inspection.duration]
            data.append(data_inspection)

        column_names = cls.extract_value_list_from_class_public_attributes(SystemTemplateInspectionsColumn)
        dataframe = pandas.DataFrame(data, columns=column_names)
        return dataframe

    @classmethod
    def _export_simulation_to_dataframe(cls, simulation:Simulation):
        data = []
        # TODO
        # If a simulation is not an instance of Monte Carlo, this code will fail 
        # (data_simulation referenced before assignment)
        if isinstance(simulation, MonteCarlo):
            column_names = [SimulationSheet.TYPE]
            column_names.extend(cls.extract_value_list_from_class_public_attributes(SimulationMonteCarloColumn))
            data_simulation = [
                simulation.SIMULATION_TYPE,
                simulation.minimum_number_of_simulations,
                simulation.maximum_number_of_simulations,
                simulation.convergence_margin,
                simulation.maximum_execution_time,
                simulation.seed,
                simulation.list_of_diagnosis,
                simulation.duration_of_simulation
            ]
        data.append(data_simulation)

        dataframe = pandas.DataFrame(data, columns=column_names)
        return dataframe

    @classmethod
    def _export_phase_to_dataframe(cls, phase_set:Set[Phase]):
        data = []
        column_names = cls.extract_value_list_from_class_public_attributes(SystemTemplatePhasesColumn)

        for phase in phase_set:
            data_phase = [phase.name,
                          phase.law.name,
                          phase.law.parameters,
                          phase.next.name,
                          phase.is_first_phase,
                          phase.next_phase_if_failure.name]
            data.append(data_phase)

        dataframe = pandas.DataFrame(data, columns=column_names)
        return dataframe

    @classmethod
    def _export_rca_to_dataframe(cls, rca_triggers_set:Set[RootCauseAnalysisTrigger]):
        data = []
        column_names = cls.extract_value_list_from_class_public_attributes(SystemTemplateRootCauseAnalysisColumn)

        for trigger in rca_triggers_set:
            data_trigger = [trigger.component_name,
                            trigger.component_status,
                            trigger.phase]
            data.append(data_trigger)

        dataframe = pandas.DataFrame(data, columns=column_names)
        return dataframe

    @classmethod
    def _export_phase_jump_to_dataframe(cls, phase_jump_trigger_set:Set[PhaseJumpTrigger]):
        data = []
        column_names = cls.extract_value_list_from_class_public_attributes(SystemTemplatePhaseJumpColumn)

        for phase_jump in phase_jump_trigger_set:
            data_phase_jump = [
                phase_jump.component_name,
                phase_jump.component_status,
                phase_jump.from_phase.name,
                phase_jump.to_phase.name
            ]
            data.append(data_phase_jump)

        dataframe = pandas.DataFrame(data, columns=column_names)
        return dataframe



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from __future__ import annotations

from typing import TYPE_CHECKING, List, Set

from availsim4core.src.discrete_event_simulation.event.b_event.b_event import BEventPriority
from availsim4core.src.discrete_event_simulation.event.b_event.held_event.end_holding_event import EndHoldingEvent

if TYPE_CHECKING:
    from availsim4core.src.context.system.component_tree.basic import Basic

from availsim4core.src.context.context import Context
from availsim4core.src.context.system.component_tree.status import Status
from availsim4core.src.context.system.failure_mode import FailureMode
from availsim4core.src.discrete_event_simulation.event.b_event.basic_b_event import BasicBEvent
from availsim4core.src.discrete_event_simulation.event.b_event.failure_event.blind_failure_event import \
    BlindFailureEvent
from availsim4core.src.discrete_event_simulation.event.b_event.failure_event.detectable_failure_event import \
    DetectableFailureEvent
from availsim4core.src.discrete_event_simulation.event.b_event.repair_event.end_repairing_event import EndRepairingEvent
from availsim4core.src.discrete_event_simulation.event.b_event.repair_event.start_repairing_event import \
    StartRepairingEvent
from availsim4core.src.discrete_event_simulation.event.c_event.c_event import CEvent, CEventPriority
from availsim4core.src.discrete_event_simulation.event.c_event.inspection_event.order_end_inspection_event import \
    OrderEndInspectionEvent


class StartInspectionEvent(BasicBEvent):
    """
    Defines the Start of an inspection event.
    Attributes:
        - absolute_occurrence_time: date when the inspection event will be started.
        - context
        - basic: component to inspect.
        - failure_mode: failure mode which triggered this inspection event.
    """
    __slots__ = 'failure_mode'

    def __init__(self,
                 absolute_occurrence_time: float,
                 context: Context,
                 basic: Basic,
                 failure_mode: FailureMode):
        super().__init__(absolute_occurrence_time, context, basic,
                         priority=BEventPriority.START_INSPECTION_EVENT)
        self.failure_mode = failure_mode

    def __hash__(self):
        return hash((type(self), self.absolute_occurrence_time, self.priority, self.basic, self.failure_mode))

    def __eq__(self, other):
        if not isinstance(other, StartInspectionEvent):
            return NotImplemented
        return super().__eq__(other) and \
               self.failure_mode == other.failure_mode

    def __str__(self):
        return (f"type ::{type(self)} "
                f"basic ::{self.basic.name}, "
                f"lid ::{self.basic.local_id}, "
                f"gid ::{self.basic.global_id}, "
                f"at t ::{self.absolute_occurrence_time}, "
                f"FailureMode ::{self.failure_mode}\n")

    def execute(self):
        return self.basic.update_status(Status.INSPECTION,
                                        self.absolute_occurrence_time,
                                        self.context.phase_manager.current_phase,  # or phase of the failure mode?
                                        "INSPECTION",
                                        self.context)

    @staticmethod
    def _b_events_to_be_cleaned() -> List:
        return [DetectableFailureEvent, BlindFailureEvent,
                StartRepairingEvent, EndRepairingEvent,
                EndHoldingEvent]

    def generate_c_event(self, **kwargs) -> Set[CEvent]:
        return {
            OrderEndInspectionEvent(
                priority=CEventPriority.ORDER_END_INSPECTION_EVENT,
                context=self.context,
                component=self.basic,
                event=self,
                failure_mode=self.failure_mode
            )
        }



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from availsim4core.src.context.system.system_template import SystemTemplate
from availsim4core.src.simulation.simulation import Simulation


class Analysis:
    """
    Domain object of which represents the an Analysis to be performed.
    Contains both the {SystemTemplate} and {simulation}.
    """
    __slots__ = 'id', 'system_template', 'simulation'

    def __init__(self,
                 identifier: int,
                 system_template: SystemTemplate,
                 simulation: Simulation):
        self.id = identifier
        self.system_template = system_template
        self.simulation = simulation

    def __eq__(self, other):
        return self.id == other.id \
               and self.system_template == other.system_template \
               and self.simulation == other.simulation

    def __str__(self):
        return f"id={self.id}, system_template={self.system_template}, simulation={self.simulation}"



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import logging
from typing import Set

from availsim4core.src.context.phase.phase import Phase
from availsim4core.src.context.system.failure import Failure, FailureType
from availsim4core.src.context.system.inspection import Inspection
from availsim4core.src.context.system.probability_law.probability_law import ProbabilityLaw
import numpy


class FailureModeError(Exception):
    pass


class FailureMode:
    """
    Defines the failure attributes based on a probability law.
    Also defines the repair attributes related to this failure based on a probability law.
    # TODO : Refactoring of this class introducing dedicated classes for the failure / repair / inspection :
    cf https://gitlab.cern.ch/availsim4/availsim4core/-/issues/30
    """
    __slots__ = 'name', 'failure_law', 'repair_law', 'failure', 'held_before_repair_phase_set', 'inspection', \
                'phase_set', 'failure_mode_next_phase_if_failure', 'phase_change_trigger', 'held_after_repair_phase_set', \
                'uniform_samples_for_quasi_monte_carlo', 'sample_generation_time'

    def __init__(self,
                 name: str,
                 failure_law: ProbabilityLaw,
                 repair_law: ProbabilityLaw,
                 failure: Failure,
                 held_before_repair_phase_set: Set[Phase],
                 inspection: Inspection,
                 phase_set: Set[Phase],
                 failure_mode_next_phase_if_failure: Phase,
                 phase_change_trigger: str,
                 held_after_repair_phase_set: Set[Phase]):

        if failure.type_of_failure == FailureType.BLIND and phase_change_trigger == 'AFTER_REPAIR':
            message = f"Wrong combination of type_of_failure ({failure}) " \
                      f"and phase_change_timing ({phase_change_trigger}) " \
                      f"for the failure mode named {name}"
            logging.exception(message)
            raise FailureModeError(message)

        self.name = name
        self.failure_law = failure_law
        self.repair_law = repair_law
        self.failure = failure
        self.held_before_repair_phase_set = held_before_repair_phase_set
        self.inspection = inspection
        self.phase_set = phase_set
        self.failure_mode_next_phase_if_failure = failure_mode_next_phase_if_failure
        self.phase_change_trigger = phase_change_trigger
        self.held_after_repair_phase_set = held_after_repair_phase_set
        self.uniform_samples_for_quasi_monte_carlo = []
        self.sample_generation_time = 0.0

    def __hash__(self):
        return hash((type(self), self.name))

    def __str__(self):
        return (f"{self.name} :: "
                f"{self.failure_law} / "
                f"{self.repair_law} / "
                f"{self.failure} / "
                f"{[phase.name for phase in self.held_before_repair_phase_set]} / "
                f"{self.inspection} / "
                f"{[phase.name for phase in self.phase_set]} / "
                f"{self.phase_change_trigger} /"
                f"{[phase.name for phase in self.held_after_repair_phase_set]}")

    def __eq__(self, other):
        if not isinstance(other, FailureMode):
            return NotImplemented
        return self.name == other.name

    def set_uniform_samples_for_quasi_monte_carlo(self, uniform_samples_for_quasi_monte_carlo: numpy.ndarray):
        self.uniform_samples_for_quasi_monte_carlo = list(uniform_samples_for_quasi_monte_carlo)


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from __future__ import annotations

from typing import TYPE_CHECKING

from availsim4core.src.discrete_event_simulation.event.b_event.b_event import BEventPriority

if TYPE_CHECKING:
    from availsim4core.src.context.system.component_tree.basic import Basic

from availsim4core.src.context.context import Context
from availsim4core.src.context.system.component_tree.status import Status
from availsim4core.src.discrete_event_simulation.event.b_event.basic_b_event import BasicBEvent
from availsim4core.src.discrete_event_simulation.event.event import Event


class StartRepairingEvent(BasicBEvent):
    """
    Defines a Repair event.
    Attributes:
        - absolute_occurrence_time: date when the repair event will be started.
        - component: associate component to repair.
        - event: event which triggered this repair event.
    """
    __slots__ = 'event'

    def __init__(self,
                 absolute_occurrence_time: float,
                 context: Context,
                 basic: Basic,
                 event: Event):
        super().__init__(absolute_occurrence_time, context, basic,
                         BEventPriority.START_REPAIRING_EVENT)
        self.event = event

    def __hash__(self):
        return hash((type(self), self.absolute_occurrence_time, self.priority, self.basic, self.event))

    def __eq__(self, other):
        if not isinstance(other, StartRepairingEvent):
            return NotImplemented
        return super().__eq__(other) and \
               self.event == other.event

    def __str__(self):
        return (f"type ::{type(self)} "
                f"basic ::{self.basic.name}, "
                f"lid ::{self.basic.local_id}, "
                f"gid ::{self.basic.global_id}, "
                f"at t ::{self.absolute_occurrence_time} \n")

    def execute(self):
        return self.basic.update_status(Status.UNDER_REPAIR,
                                        self.absolute_occurrence_time,
                                        self.context.phase_manager.current_phase,  # or phase of the failure mode?
                                        f"UNDER REPAIR - component {self.basic.name}_{self.basic.local_id}_{self.basic.global_id}",
                                        self.context)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import glob
import os
import random
import shutil
import unittest

from availsim4core import main
from availsim4core.test.E2E.E2E_utils import Result


class test_E2E_multiple_failure_modes(unittest.TestCase):

    expected_result_scenario_1 = [
        Result(95.5, "ROOT_0_1","*","RUNNING","_MEAN_DURATION"),
        Result(2, "A_1_5", "*", "RUNNING", "_MEAN_OCCURRENCES"),
        Result(1, "A_1_5", "*", "FAILED", "_MEAN_OCCURRENCES"),
        Result(1, "B_0_11", "*", "RUNNING", "_MEAN_OCCURRENCES"),
        Result(2, "D_0_3", "*", "RUNNING", "_MEAN_OCCURRENCES"),
        Result(1, "D_0_3", "*", "FAILED", "_MEAN_OCCURRENCES"),
        Result(2, "D_0_6", "*", "RUNNING", "_MEAN_OCCURRENCES"),
        Result(1, "D_0_6", "*", "FAILED", "_MEAN_OCCURRENCES"),
        Result(2, "D_0_9", "*", "RUNNING", "_MEAN_OCCURRENCES"),
        Result(1, "D_0_9", "*", "FAILED", "_MEAN_OCCURRENCES"),
        Result(2, "E_0_4", "*", "RUNNING", "_MEAN_OCCURRENCES"),
        Result(1, "E_0_4", "*", "FAILED", "_MEAN_OCCURRENCES"),
        Result(2, "E_0_7", "*", "RUNNING", "_MEAN_OCCURRENCES"),
        Result(1, "E_0_7", "*", "FAILED", "_MEAN_OCCURRENCES"),
        Result(2, "E_0_10", "*", "RUNNING", "_MEAN_OCCURRENCES"),
        Result(1, "E_0_10", "*", "FAILED", "_MEAN_OCCURRENCES")
    ]
    for component_name in ["Phase","ROOT_0_1",
                           "A_0_2","D_0_3","E_0_4","A_1_5","D_0_6","E_0_7","A_2_8","D_0_9","E_0_10","B_0_11","C_0_12"]:
        expected_result_scenario_1.append(Result(100, component_name, "*", "*", "_MEAN_DURATION"))
    scenario_1 = \
        ["./availsim4core/test/E2E/input/multiple_failure_modes/simulation.xlsx",
         "./availsim4core/test/E2E/input/multiple_failure_modes/no_multiple_fm__no_phase__no_held__system.xlsx",
         None,
         1,
         expected_result_scenario_1]

    expected_result_scenario_2 = [
        Result(95.5, "ROOT_0_1", "*", "RUNNING", "_MEAN_DURATION"),
#        Result(2, "A_1_5", "*", "RUNNING", "_MEAN_OCCURRENCES"),
#        Result(1, "A_1_5", "*", "FAILED", "_MEAN_OCCURRENCES"),
#        Result(1, "B_0_11", "*", "RUNNING", "_MEAN_OCCURRENCES"),
#        Result(2, "D_0_3", "*", "RUNNING", "_MEAN_OCCURRENCES"),
#        Result(1, "D_0_3", "*", "FAILED", "_MEAN_OCCURRENCES"),
#        Result(2, "D_0_6", "*", "RUNNING", "_MEAN_OCCURRENCES"),
#        Result(1, "D_0_6", "*", "FAILED", "_MEAN_OCCURRENCES"),
#        Result(2, "D_0_9", "*", "RUNNING", "_MEAN_OCCURRENCES"),
#        Result(1, "D_0_9", "*", "FAILED", "_MEAN_OCCURRENCES"),
#        Result(2, "E_0_4", "*", "RUNNING", "_MEAN_OCCURRENCES"),
#        Result(1, "E_0_4", "*", "FAILED", "_MEAN_OCCURRENCES"),
#        Result(2, "E_0_7", "*", "RUNNING", "_MEAN_OCCURRENCES"),
#        Result(1, "E_0_7", "*", "FAILED", "_MEAN_OCCURRENCES"),
#        Result(2, "E_0_10", "*", "RUNNING", "_MEAN_OCCURRENCES"),
#        Result(1, "E_0_10", "*", "FAILED", "_MEAN_OCCURRENCES")
    ]
    for component_name in ["Phase", "ROOT_0_1",
                           "A_0_2", "D_0_3", "E_0_4", "A_1_5", "D_0_6", "E_0_7", "A_2_8", "D_0_9", "E_0_10", "B_0_11",
                           "C_0_12"]:
        expected_result_scenario_2.append(Result(100, component_name, "*", "*", "_MEAN_DURATION"))
    scenario_2 = \
        ["./availsim4core/test/E2E/input/multiple_failure_modes/simulation.xlsx",
         "./availsim4core/test/E2E/input/multiple_failure_modes/multiple_fm__no_phase__no_held__system.xlsx",
         None,
         1,
         expected_result_scenario_2]

    param_scenario = [
        scenario_1,
        #scenario_2, # need for a bug fix !
    ]

    def test_runner(self):
        # randomize the order of the test execution to detect hidden dependencies
        random.shuffle(self.param_scenario)
        for simulation_file, system_file, sensitivity_file, expected_number_file_result, expected_result_list in self.param_scenario:
            with self.subTest():
                self._runner_E2E(simulation_file,
                                 system_file,
                                 sensitivity_file,
                                 expected_number_file_result,
                                 expected_result_list)

    def _runner_E2E(self,
                    simulation_file,
                    system_file,
                    sensitivity_file,
                    expected_number_file_result,
                    expected_result_list):

        output_folder = "./availsim4core/test/E2E/output/multiple_failure_modes/"

        # Clean folder
        if os.path.exists(output_folder):
            shutil.rmtree(output_folder)

        # Run the main process
        main.start(simulation_file, system_file, sensitivity_file, output_folder)

        result_simulation_file_list = glob.glob(output_folder + "/*.xlsx")
        self.assertEqual(len(result_simulation_file_list), expected_number_file_result)

        for expected_result in expected_result_list:
            actual_result = expected_result.extract_result(result_simulation_file_list)

            if expected_result.tolerance == -1:
                self.assertEqual(expected_result.expected_result, actual_result)
            else:
                self.assertAlmostEqual(expected_result.expected_result, actual_result, places=expected_result.tolerance)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import glob
import os
import random
import shutil
import unittest

from availsim4core import main
from availsim4core.test.E2E.E2E_utils import Result


class test_E2E_phases(unittest.TestCase):
    # no phase ==> already done by all other tests without phase
    # test with phases are for now done with AND logic only, that feature is customized for the LHC model

    ###
    # one failure occurs after a given deterministic time which is smaller than the duration of the simulation but still
    # the failure never occurs because the system does not spend enough time in that phase
    expected_result_scenario_1 = [Result(11, "A2_0_11", "A", "RUNNING", "_MEAN_OCCURRENCES"),
                                  Result(0, "A2_0_11", "A", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(11, "Phase", "A", "_", "_MEAN_OCCURRENCES")]
    for component_name in ["Phase", "D_0_1", "C1_0_2", "B1_0_3", "B2_0_4", "C2_0_5", "B3_0_6", "B3_1_7", "B4_0_8",
                           "A1_0_9", "A1_1_10", "A2_0_11", "C3_0_12"]:
        expected_result_scenario_1.append(Result(100, component_name, "*", "*", "_MEAN_DURATION"))
    scenario_1 = \
        ["./availsim4core/test/E2E/input/phases/simulation.xlsx",
         "./availsim4core/test/E2E/input/phases/no_sufficient_time_in_faulty_phase_system.xlsx",
         None,
         1,
         expected_result_scenario_1]

    ###
    # ten failures occur in a given phase
    expected_result_scenario_2 = [Result(10, "A2_0_11", "A", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(11, "Phase", "A", "_", "_MEAN_OCCURRENCES")]
    for component_name in ["Phase", "D_0_1", "C1_0_2", "B1_0_3", "B2_0_4", "C2_0_5", "B3_0_6", "B3_1_7", "B4_0_8",
                           "A1_0_9", "A1_1_10", "A2_0_11", "C3_0_12"]:
        expected_result_scenario_2.append(Result(100, component_name, "*", "*", "_MEAN_DURATION"))
    scenario_2 = \
        ["./availsim4core/test/E2E/input/phases/simulation.xlsx",
         "./availsim4core/test/E2E/input/phases/sufficient_time_in_faulty_phase_system.xlsx",
         None,
         1,
         expected_result_scenario_2]

    ###
    # 100 failures occur in a given phase, same input as scenario_2 but when a failure occurs in phase A,
    # the system go back to phase A
    expected_result_scenario_2b = [Result(100, "A2_0_11", "A", "FAILED", "_MEAN_OCCURRENCES"),
                                   Result(101, "Phase", "A", "_", "_MEAN_OCCURRENCES")]
    for component_name in ["Phase", "D_0_1", "C1_0_2", "B1_0_3", "B2_0_4", "C2_0_5", "B3_0_6", "B3_1_7", "B4_0_8",
                           "A1_0_9", "A1_1_10", "A2_0_11", "C3_0_12"]:
        expected_result_scenario_2b.append(Result(100, component_name, "*", "*", "_MEAN_DURATION"))
    scenario_2b = \
        ["./availsim4core/test/E2E/input/phases/simulation.xlsx",
         "./availsim4core/test/E2E/input/phases/sufficient_time_in_faulty_phase_v2_system.xlsx",
         None,
         1,
         expected_result_scenario_2b]

    ###
    # the system is never passing in the phase where a failure could occur
    expected_result_scenario_3 = [Result(11, "A2_0_11", "B", "RUNNING", "_MEAN_OCCURRENCES"),
                                  Result(0, "Phase", "A", "_", "_MEAN_OCCURRENCES"),
                                  Result(11, "Phase", "B", "_", "_MEAN_OCCURRENCES")]
    for component_name in ["Phase", "D_0_1", "C1_0_2", "B1_0_3", "B2_0_4", "C2_0_5", "B3_0_6", "B3_1_7", "B4_0_8",
                           "A1_0_9", "A1_1_10", "A2_0_11", "C3_0_12"]:
        expected_result_scenario_3.append(Result(100, component_name, "*", "*", "_MEAN_DURATION"))
    scenario_3 = \
        ["./availsim4core/test/E2E/input/phases/simulation.xlsx",
         "./availsim4core/test/E2E/input/phases/never_passing_in_faulty_phase_system.xlsx",
         None,
         1,
         expected_result_scenario_3]

    ###
    # system is passing only once in the phase where a failure occurs
    expected_result_scenario_4 = [Result(1, "A2_0_11", "A", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(1, "Phase", "A", "_", "_MEAN_OCCURRENCES")]
    for component_name in ["Phase", "D_0_1", "C1_0_2", "B1_0_3", "B2_0_4", "C2_0_5", "B3_0_6", "B3_1_7", "B4_0_8",
                           "A1_0_9", "A1_1_10", "A2_0_11", "C3_0_12"]:
        expected_result_scenario_4.append(Result(100, component_name, "*", "*", "_MEAN_DURATION"))
    scenario_4 = \
        ["./availsim4core/test/E2E/input/phases/simulation.xlsx",
         "./availsim4core/test/E2E/input/phases/passing_once_in_faulty_phase_system.xlsx",
         None,
         1,
         expected_result_scenario_4]

    ###
    # no phase, 50 failures
    expected_result_scenario_5 = [Result(50, "ROOT_0_1", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(50, "B_0_11", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(50, "ROOT_0_1", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  Result(50, "B_0_11", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  Result(51, "ROOT_0_1", "*", "RUNNING", "_MEAN_OCCURRENCES"),
                                  Result(51, "B_0_11", "*", "RUNNING", "_MEAN_OCCURRENCES")
                                  ]

    for component_name in ["Phase", "ROOT_0_1", "A_0_2", "D_0_3", "E_0_4", "A_1_5", "D_0_6", "E_0_7", "A_2_8", "D_0_9",
                           "E_0_10","B_0_11", "C_0_12"]:
        expected_result_scenario_5.append(Result(100, component_name, "*", "*", "_MEAN_DURATION"))

    for component_name in ["A_0_2", "D_0_3", "E_0_4", "A_1_5", "D_0_6", "E_0_7", "A_2_8","D_0_9",
                               "E_0_10", "C_0_12"]:
        expected_result_scenario_5.append(Result(1, component_name, "*", "RUNNING", "_MEAN_OCCURRENCES"))

    scenario_5 = \
        ["./availsim4core/test/E2E/input/phases/simulation.xlsx",
         "./availsim4core/test/E2E/input/phases/version1_no_phase_system.xlsx",
         None,
         1,
         expected_result_scenario_5]

    ###
    # scenario using phases but the unique failure is present in each phase, should be equivalent to scenario 5
    expected_result_scenario_6 = [Result(50, "ROOT_0_1", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(50, "B_0_11", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(50, "ROOT_0_1", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  Result(50, "B_0_11", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  #Result(126, "ROOT_0_1", "*", "RUNNING", "_MEAN_OCCURRENCES"),
                                  #Result(126, "B_0_11", "*", "RUNNING", "_MEAN_OCCURRENCES")
                                  ]
    # 126 instead of 51 like in scenario 5 because of the many times the system enters new phases
    # the 126 is not deterministic TODO: DETERMINISTIC CHECK
    for component_name in ["Phase", "ROOT_0_1", "A_0_2", "D_0_3", "E_0_4", "A_1_5", "D_0_6", "E_0_7", "A_2_8", "D_0_9",
                           "E_0_10","B_0_11", "C_0_12"]:
        expected_result_scenario_6.append(Result(100, component_name, "*", "*", "_MEAN_DURATION"))
    #for component_name in ["A_0_2", "D_0_3", "E_0_4", "A_1_5", "D_0_6", "E_0_7", "A_2_8","D_0_9",
    #                           "E_0_10", "C_0_12"]:
    #    expected_result_scenario_6.append(Result(100, component_name, "*", "RUNNING", "_MEAN_OCCURRENCES"))
    # TODO: DETERMINISTIC CHECK

    scenario_6 = \
        ["./availsim4core/test/E2E/input/phases/simulation.xlsx",
         "./availsim4core/test/E2E/input/phases/version2_failure_in_each_phase_system.xlsx",
         None,
         1,
         expected_result_scenario_6]

    ###
    # scenario using phases but the unique failure is present in each phase, should be equivalent to scenario 5 and 6
    # but with a different transition of phase when a failure occurs
    expected_result_scenario_6b = [Result(50, "ROOT_0_1", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(50, "B_0_11", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(50, "ROOT_0_1", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  Result(50, "B_0_11", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  # Result(126, "ROOT_0_1", "*", "RUNNING", "_MEAN_OCCURRENCES"),
                                  # Result(126, "B_0_11", "*", "RUNNING", "_MEAN_OCCURRENCES")
                                  ]
    # 126 instead of 51 like in scenario 5 because of the many times the system enters new phases
    # also, 126 is not deterministic TODO: DETERMINISTIC CHECK
    for component_name in ["Phase", "ROOT_0_1", "A_0_2", "D_0_3", "E_0_4", "A_1_5", "D_0_6", "E_0_7", "A_2_8", "D_0_9",
                           "E_0_10", "B_0_11", "C_0_12"]:
        expected_result_scenario_6b.append(Result(100, component_name, "*", "*", "_MEAN_DURATION"))
    #for component_name in ["A_0_2", "D_0_3", "E_0_4", "A_1_5", "D_0_6", "E_0_7", "A_2_8","D_0_9",
    #                           "E_0_10", "C_0_12"]:
    #    expected_result_scenario_6.append(Result(100, component_name, "*", "RUNNING", "_MEAN_OCCURRENCES"))
    # TODO: DETERMINISTIC CHECK
    scenario_6b = \
        ["./availsim4core/test/E2E/input/phases/simulation.xlsx",
         "./availsim4core/test/E2E/input/phases/version2_failure_in_each_phase_v2_system.xlsx",
         None,
         1,
         expected_result_scenario_6b]

    ###
    # scenario using phases but the unique failure is present in each phase, should be equivalent to scenario 5 and 6 and 6b
    # but with a different transition of phase when a failure occurs
    expected_result_scenario_6c = [Result(50, "ROOT_0_1", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(50, "B_0_11", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(50, "ROOT_0_1", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  Result(50, "B_0_11", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  # Result(126, "ROOT_0_1", "*", "RUNNING", "_MEAN_OCCURRENCES"),
                                  # Result(126, "B_0_11", "*", "RUNNING", "_MEAN_OCCURRENCES")
                                  ]
    # 126 instead of 51 like in scenario 5 because of the many times the system enters new phases
    # also, 126 is not deterministic TODO: DETERMINISTIC CHECK
    for component_name in ["Phase", "ROOT_0_1", "A_0_2", "D_0_3", "E_0_4", "A_1_5", "D_0_6", "E_0_7", "A_2_8", "D_0_9",
                           "E_0_10", "B_0_11", "C_0_12"]:
        expected_result_scenario_6c.append(Result(100, component_name, "*", "*", "_MEAN_DURATION"))
    #for component_name in ["A_0_2", "D_0_3", "E_0_4", "A_1_5", "D_0_6", "E_0_7", "A_2_8","D_0_9",
    #                           "E_0_10", "C_0_12"]:
    #    expected_result_scenario_6c.append(Result(100, component_name, "*", "RUNNING", "_MEAN_OCCURRENCES"))
    # TODO: DETERMINISTIC CHECK
    scenario_6c = \
        ["./availsim4core/test/E2E/input/phases/simulation.xlsx",
         "./availsim4core/test/E2E/input/phases/version2_failure_in_each_phase_v3_system.xlsx",
         None,
         1,
         expected_result_scenario_6c]

    ###
    # scenario using phases but the unique failure is present in each phase using 2 different failure modes in 2 different basics,
    # should be equivalent to scenario 5
    expected_result_scenario_7 = [Result(50, "ROOT_0_1", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(26, "BFM12_0_11", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(24, "BFM34_0_12", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(50, "ROOT_0_1", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  Result(26, "BFM12_0_11", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  Result(24, "BFM34_0_12", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES")]
    for component_name in ["Phase", "ROOT_0_1", "A_0_2", "D_0_3", "E_0_4", "A_1_5", "D_0_6", "E_0_7", "A_2_8", "D_0_9",
                           "E_0_10", "BFM12_0_11", "BFM34_0_12", "C_0_13"]:
        expected_result_scenario_7.append(Result(100, component_name, "*", "*", "_MEAN_DURATION"))
    scenario_7 = \
        ["./availsim4core/test/E2E/input/phases/simulation.xlsx",
         "./availsim4core/test/E2E/input/phases/version3_failure_in_each_phase_2_failure_modes_system.xlsx",
         None,
         1,
         expected_result_scenario_7]

    ###
    # scenario using phases but the unique failure is present in each phase using 2 different failure mode, should be equivalent to scenario 5
    # and scenario 7 but with a different transition scheme when a fault occurs
    expected_result_scenario_7b = [Result(50, "ROOT_0_1", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                   Result(25, "BFM12_0_11", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                   Result(25, "BFM34_0_12", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(50, "ROOT_0_1", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  Result(25, "BFM12_0_11", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  Result(25, "BFM34_0_12", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES")]
    for component_name in ["Phase", "ROOT_0_1", "A_0_2", "D_0_3", "E_0_4", "A_1_5", "D_0_6", "E_0_7", "A_2_8", "D_0_9",
                           "E_0_10", "BFM12_0_11", "BFM34_0_12", "C_0_13"]:
        expected_result_scenario_7b.append(Result(100, component_name, "*", "*", "_MEAN_DURATION"))
    scenario_7b = \
        ["./availsim4core/test/E2E/input/phases/simulation.xlsx",
         "./availsim4core/test/E2E/input/phases/version3_failure_in_each_phase_2_failure_modes_v2_system.xlsx",
         None,
         1,
         expected_result_scenario_7b]

    # TODO: in scenario_7 and scenario_7b, the durations of the phase are multiples of the timetofail+timetorepair
    # to avoid any complication, it should be changed for the test to be even more general

    ###
    expected_result_scenario_8 = [Result(3, "B_0_11", "*", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(1, "Phase", "P1", "_", "_MEAN_OCCURRENCES"),
                                  Result(2, "Phase", "P2", "_", "_MEAN_OCCURRENCES"),
                                  Result(2, "Phase", "P3", "_", "_MEAN_OCCURRENCES")]
    for component_name in ["Phase", "ROOT_0_1", "A_0_2", "D_0_3", "E_0_4", "A_1_5", "D_0_6", "E_0_7", "A_2_8", "D_0_9",
                           "E_0_10",
                           "B_0_11", "C_0_12"]:
        expected_result_scenario_8.append(Result(100, component_name, "*", "*", "_MEAN_DURATION"))
    scenario_8 = \
        ["./availsim4core/test/E2E/input/phases/simulation.xlsx",
         "./availsim4core/test/E2E/input/phases/failure_in_each_Phase_but_change_phase_before_failure_system.xlsx",
         None,
         1,
         expected_result_scenario_8]

    ###
    # the system never fails but is passing in different phases
    expected_result_scenario_9 = [Result(15, "Phase", "P1", "*", "_MEAN_OCCURRENCES"),
                                  Result(15, "Phase", "P2", "*", "_MEAN_OCCURRENCES"),
                                  Result(14, "Phase", "P3", "*", "_MEAN_OCCURRENCES"),
                                  Result(14, "Phase", "P4", "*", "_MEAN_OCCURRENCES")]
    for component_name in ["Phase",
                           "ROOT_0_1", "A_0_2", "D_0_3", "E_0_4", "A_1_5", "D_0_6", "E_0_7", "A_2_8", "D_0_9", "E_0_10",
                           "B_0_11", "C_0_12"]:
        expected_result_scenario_9.append(Result(100, component_name, "*", "*", "_MEAN_DURATION", tolerance=6))
        dict_of_phase_duration = {
            "P1": 15,
            "P2": 28.86,
            "P3": 56,
            "P4": 0.14
        }
        for phase, value in dict_of_phase_duration.items():
            expected_result_scenario_9.append(Result(value, component_name, phase, "*", "_MEAN_DURATION", tolerance=6))
    scenario_9 = \
        ["./availsim4core/test/E2E/input/phases/simulation.xlsx",
         "./availsim4core/test/E2E/input/phases/version4_phases_but_no_failures_system.xlsx",
         None,
         1,
         expected_result_scenario_9]

    ###
    # a failure force the system into a specific phase
    expected_result_scenario_10 = [
        Result(10, "Phase", "P1", "*", "_MEAN_OCCURRENCES"),
        Result(10, "Phase", "P2", "*", "_MEAN_OCCURRENCES"),
        Result(10, "Phase", "P3", "*", "_MEAN_OCCURRENCES"),
        Result(6, "Phase", "P4", "*", "_MEAN_OCCURRENCES"),
        Result(3, "Phase", "CYCLING", "*", "_MEAN_OCCURRENCES"),
        Result(10, "Phase", "P1", "*", "_MEAN_DURATION", tolerance=6),
        Result(20, "Phase", "P2", "*", "_MEAN_DURATION", tolerance=6),
        Result(39.94, "Phase", "P3", "*", "_MEAN_DURATION", tolerance=6),
        Result(0.06, "Phase", "P4", "*", "_MEAN_DURATION", tolerance=6),
        Result(30, "Phase", "CYCLING", "*", "_MEAN_DURATION", tolerance=6),
    ]
    scenario_10 = \
        ["./availsim4core/test/E2E/input/phases/simulation.xlsx",
         "./availsim4core/test/E2E/input/phases/version5_phase_specific_to_failure_mode_system.xlsx",
         None,
         1,
         expected_result_scenario_10]


    ###
    # a failure forces the system into a specific phase
    # but another fault starts before the "failure needing a specific phase"
    # and that failure is repaired before the "failure needing a specific phase"
    expected_result_scenario_10b = [
        Result(10, "Phase", "P1", "*", "_MEAN_OCCURRENCES"),
        Result(10, "Phase", "P2", "*", "_MEAN_OCCURRENCES"),
        Result(10, "Phase", "P3", "*", "_MEAN_OCCURRENCES"),
        Result(6, "Phase", "P4", "*", "_MEAN_OCCURRENCES"),
        Result(3, "Phase", "CYCLING", "*", "_MEAN_OCCURRENCES"),
        Result(10, "Phase", "P1", "*", "_MEAN_DURATION", tolerance=6),
        Result(20, "Phase", "P2", "*", "_MEAN_DURATION", tolerance=6),
        Result(39.94, "Phase", "P3", "*", "_MEAN_DURATION", tolerance=6),
        Result(0.06, "Phase", "P4", "*", "_MEAN_DURATION", tolerance=6),
        Result(30, "Phase", "CYCLING", "*", "_MEAN_DURATION", tolerance=6),
    ]
    scenario_10b = \
        ["./availsim4core/test/E2E/input/phases/simulation.xlsx",
         "./availsim4core/test/E2E/input/phases/version5_phaseSpecificToFailureMode_parallelFault1_system.xlsx",
         None,
         1,
         expected_result_scenario_10b]

    ###
    # a failure forces the system into a specific phase
    # but another fault starts before the "failure needing a specific phase"
    # and that failure is repaired after the "failure needing a specific phase"
    # (this only applies to the first occurrences of those failures, because after this occurrence
    # everything gets out of sync as ttf+ttr = 12.5 for the failure needing a specific phase while
    # it's 12.5001 for the other failure)
    expected_result_scenario_10c = [
        Result(14, "Phase", "P1", "*", "_MEAN_OCCURRENCES"),
        Result(14, "Phase", "P2", "*", "_MEAN_OCCURRENCES"),
        Result(14, "Phase", "P3", "*", "_MEAN_OCCURRENCES"),
        Result(13, "Phase", "P4", "*", "_MEAN_OCCURRENCES"),
        Result(0, "Phase", "CYCLING", "*", "_MEAN_OCCURRENCES"),
        Result(14, "Phase", "P1", "*", "_MEAN_DURATION", tolerance=6),
        Result(28, "Phase", "P2", "*", "_MEAN_DURATION", tolerance=6),
        Result(57.87, "Phase", "P3", "*", "_MEAN_DURATION", tolerance=6),
        Result(0.13, "Phase", "P4", "*", "_MEAN_DURATION", tolerance=6),
        Result(0, "Phase", "CYCLING", "*", "_MEAN_DURATION", tolerance=6),
    ]
    scenario_10c = \
        ["./availsim4core/test/E2E/input/phases/simulation.xlsx",
         "./availsim4core/test/E2E/input/phases/version5_phaseSpecificToFailureMode_parallelFault2_system.xlsx",
         None,
         1,
         expected_result_scenario_10c]

    ###
    # a failure forces the system into a specific phase
    # but another fault starts after the "failure needing a specific phase"
    # and that failure is repaired before the "failure needing a specific phase"
    expected_result_scenario_10d = [
        Result(10, "Phase", "P1", "*", "_MEAN_OCCURRENCES"),
        Result(10, "Phase", "P2", "*", "_MEAN_OCCURRENCES"),
        Result(10, "Phase", "P3", "*", "_MEAN_OCCURRENCES"),
        Result(6, "Phase", "P4", "*", "_MEAN_OCCURRENCES"),
        Result(3, "Phase", "CYCLING", "*", "_MEAN_OCCURRENCES"),
        Result(10, "Phase", "P1", "*", "_MEAN_DURATION", tolerance=6),
        Result(20, "Phase", "P2", "*", "_MEAN_DURATION", tolerance=6),
        Result(39.94, "Phase", "P3", "*", "_MEAN_DURATION", tolerance=6),
        Result(0.06, "Phase", "P4", "*", "_MEAN_DURATION", tolerance=6),
        Result(30, "Phase", "CYCLING", "*", "_MEAN_DURATION", tolerance=6),
    ]
    scenario_10d = \
        ["./availsim4core/test/E2E/input/phases/simulation.xlsx",
         "./availsim4core/test/E2E/input/phases/version5_phaseSpecificToFailureMode_parallelFault3_system.xlsx",
         None,
         1,
         expected_result_scenario_10d]

    ###
    # a failure forces the system into a specific phase
    # but another fault starts before the "failure needing a specific phase"
    # and that failure is repaired before the "failure needing a specific phase"
    expected_result_scenario_10e = [
        Result(10, "Phase", "P1", "*", "_MEAN_OCCURRENCES"),
        Result(10, "Phase", "P2", "*", "_MEAN_OCCURRENCES"),
        Result(10, "Phase", "P3", "*", "_MEAN_OCCURRENCES"),
        Result(6, "Phase", "P4", "*", "_MEAN_OCCURRENCES"),
        Result(3, "Phase", "CYCLING", "*", "_MEAN_OCCURRENCES"),
        Result(10, "Phase", "P1", "*", "_MEAN_DURATION", tolerance=6),
        Result(20, "Phase", "P2", "*", "_MEAN_DURATION", tolerance=6),
        Result(39.94, "Phase", "P3", "*", "_MEAN_DURATION", tolerance=6),
        Result(0.06, "Phase", "P4", "*", "_MEAN_DURATION", tolerance=6),
        Result(30, "Phase", "CYCLING", "*", "_MEAN_DURATION", tolerance=6),
    ]
    scenario_10e = \
        ["./availsim4core/test/E2E/input/phases/simulation.xlsx",
         "./availsim4core/test/E2E/input/phases/version5_phaseSpecificToFailureMode_parallelFault4_system.xlsx",
         None,
         1,
         expected_result_scenario_10e]

    ###
    # testing the option to change phase after a repair or after a failure
    # baseline = after a repair
    # in this simulation, the repair is too long to finish, the phase never changes
    expected_result_scenario_11a = [
        Result(1, "Phase", "P1", "*", "_MEAN_OCCURRENCES"),
        Result(0, "Phase", "P2", "*", "_MEAN_OCCURRENCES"),
        Result(100, "Phase", "P1", "*", "_MEAN_DURATION", tolerance=6),
        Result(0, "Phase", "P2", "*", "_MEAN_DURATION", tolerance=6),
    ]
    scenario_11a = \
        ["./availsim4core/test/E2E/input/phases/simulation.xlsx",
         "./availsim4core/test/E2E/input/phases/option_on_timing__after_repair__system.xlsx",
         None,
         1,
         expected_result_scenario_11a]

    ###
    # testing the option to change phase after a repair or after a failure
    # new possibility = after a failure
    # in this simulation, the repair is too long to finish, still the phase changes after the failure
    expected_result_scenario_11b = [
        Result(1, "Phase", "P1", "*", "_MEAN_OCCURRENCES"),
        Result(1, "Phase", "P2", "*", "_MEAN_OCCURRENCES"),
        Result(50, "Phase", "P1", "*", "_MEAN_DURATION", tolerance=6),
        Result(50, "Phase", "P2", "*", "_MEAN_DURATION", tolerance=6),
    ]
    scenario_11b = \
        ["./availsim4core/test/E2E/input/phases/simulation.xlsx",
         "./availsim4core/test/E2E/input/phases/option_on_timing__after_failure__system.xlsx",
         None,
         1,
         expected_result_scenario_11b]

    ###
    # testing the option to change phase after a repair or after a failure
    # baseline = after a repair
    # in this simulation, the repair is not too long to finish, the phase changes
    expected_result_scenario_11c = [
        Result(1, "Phase", "P1", "*", "_MEAN_OCCURRENCES"),
        Result(1, "Phase", "P2", "*", "_MEAN_OCCURRENCES"),
        Result(80, "Phase", "P1", "*", "_MEAN_DURATION", tolerance=6),
        Result(20, "Phase", "P2", "*", "_MEAN_DURATION", tolerance=6),
    ]
    scenario_11c = \
        ["./availsim4core/test/E2E/input/phases/simulation.xlsx",
         "./availsim4core/test/E2E/input/phases/option_on_timing_v2__after_repair__system.xlsx",
         None,
         1,
         expected_result_scenario_11c]

    ###
    # testing the option to change phase after a repair or after a failure
    # new = never
    # in this simulation, the repair is not too long to finish but the phase never changes
    expected_result_scenario_11d = [
        Result(1, "Phase", "P1", "*", "_MEAN_OCCURRENCES"),
        Result(0, "Phase", "P2", "*", "_MEAN_OCCURRENCES"),
        Result(100, "Phase", "P1", "*", "_MEAN_DURATION", tolerance=6),
        Result(0, "Phase", "P2", "*", "_MEAN_DURATION", tolerance=6),
    ]
    scenario_11d = \
        ["./availsim4core/test/E2E/input/phases/simulation.xlsx",
         "./availsim4core/test/E2E/input/phases/option_on_timing_v2__never__system.xlsx",
         None,
         1,
         expected_result_scenario_11d]

    ###
    # Test checking if postponing an event is done correctly.
    # One failure mode is define in the system, this failure mode is only valid in one of the two phases,
    # thus the failure it generates has to be postponed.
    expected_result_scenario_12_v1 = [Result(1, "ROOT_0_1", "P1", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(1, "B_0_11", "P1", "FAILED", "_MEAN_OCCURRENCES"),
                                  #Result(2, "ROOT_0_1", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"), # TODO: check deterministic
                                  #Result(2, "B_0_11", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  Result(3, "ROOT_0_1", "*", "RUNNING", "_MEAN_OCCURRENCES"),
                                  Result(3, "B_0_11", "*", "RUNNING", "_MEAN_OCCURRENCES")
                                  ]
    for component_name in ["Phase", "ROOT_0_1", "A_0_2", "D_0_3", "E_0_4", "A_1_5", "D_0_6", "E_0_7", "A_2_8", "D_0_9",
                           "E_0_10","B_0_11", "C_0_12"]:
        expected_result_scenario_12_v1.append(Result(100, component_name, "*", "*", "_MEAN_DURATION"))
    for component_name in ["A_0_2", "D_0_3", "E_0_4", "A_1_5", "D_0_6", "E_0_7", "A_2_8","D_0_9",
                               "E_0_10", "C_0_12"]:
        expected_result_scenario_12_v1.append(Result(3, component_name, "*", "RUNNING", "_MEAN_OCCURRENCES"))
    scenario_12_v1 = \
        ["./availsim4core/test/E2E/input/phases/simulation.xlsx",
         "./availsim4core/test/E2E/input/phases/version4_check_postpone_v1_system.xlsx",
         None,
         1,
         expected_result_scenario_12_v1]

    ###
    # Test checking if postponing an event is done correctly.
    # One failure mode is defined in the system, this failure mode is valid in both of the two phases
    expected_result_scenario_12_v2 = [Result(2, "ROOT_0_1", "P1", "FAILED", "_MEAN_OCCURRENCES"),
                                  Result(2, "B_0_11", "P1", "FAILED", "_MEAN_OCCURRENCES"),
                                  #Result(2, "ROOT_0_1", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"), # TODO: check deterministic
                                  #Result(2, "B_0_11", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                  Result(4, "ROOT_0_1", "*", "RUNNING", "_MEAN_OCCURRENCES"),
                                  Result(4, "B_0_11", "*", "RUNNING", "_MEAN_OCCURRENCES")
                                  ]
    for component_name in ["Phase", "ROOT_0_1", "A_0_2", "D_0_3", "E_0_4", "A_1_5", "D_0_6", "E_0_7", "A_2_8", "D_0_9",
                           "E_0_10","B_0_11", "C_0_12"]:
        expected_result_scenario_12_v2.append(Result(100, component_name, "*", "*", "_MEAN_DURATION"))
    for component_name in ["A_0_2", "D_0_3", "E_0_4", "A_1_5", "D_0_6", "E_0_7", "A_2_8","D_0_9",
                               "E_0_10", "C_0_12"]:
        expected_result_scenario_12_v2.append(Result(4, component_name, "*", "RUNNING", "_MEAN_OCCURRENCES"))
    scenario_12_v2 = \
        ["./availsim4core/test/E2E/input/phases/simulation.xlsx",
         "./availsim4core/test/E2E/input/phases/version4_check_postpone_v2_system.xlsx",
         None,
         1,
         expected_result_scenario_12_v2]

    ###
    # Test checking if postponing an event is done correctly.
    # One failure mode is defined in the system, this failure mode is valid in both of the two phases
    expected_result_scenario_12_v3 = [Result(2, "ROOT_0_1", "P1", "FAILED", "_MEAN_OCCURRENCES"),
                                      Result(2, "B_0_11", "P1", "FAILED", "_MEAN_OCCURRENCES"),
                                      # Result(2, "ROOT_0_1", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"), # TODO: check deterministic
                                      # Result(2, "B_0_11", "*", "UNDER_REPAIR", "_MEAN_OCCURRENCES"),
                                      Result(4, "ROOT_0_1", "*", "RUNNING", "_MEAN_OCCURRENCES"),
                                      Result(4, "B_0_11", "*", "RUNNING", "_MEAN_OCCURRENCES")
                                      ]
    for component_name in ["Phase", "ROOT_0_1", "A_0_2", "D_0_3", "E_0_4", "A_1_5", "D_0_6", "E_0_7", "A_2_8", "D_0_9",
                           "E_0_10", "B_0_11", "C_0_12"]:
        expected_result_scenario_12_v3.append(Result(100, component_name, "*", "*", "_MEAN_DURATION"))
    for component_name in ["A_0_2", "D_0_3", "E_0_4", "A_1_5", "D_0_6", "E_0_7", "A_2_8", "D_0_9",
                           "E_0_10", "C_0_12"]:
        expected_result_scenario_12_v3.append(Result(4, component_name, "*", "RUNNING", "_MEAN_OCCURRENCES"))
    scenario_12_v3 = \
        ["./availsim4core/test/E2E/input/phases/simulation.xlsx",
         "./availsim4core/test/E2E/input/phases/version4_check_postpone_v3_system.xlsx",
         None,
         1,
         expected_result_scenario_12_v3]

    param_scenario = [
        scenario_1,
        scenario_2,
        scenario_2b,
        scenario_3,
        scenario_4,
        scenario_5,
        scenario_6,
        scenario_6b,
        scenario_6c,
        scenario_7,
        scenario_7b,
        scenario_8,
        scenario_9,
        scenario_10,
        scenario_10b,
        scenario_10c,
        scenario_10d,
        scenario_10e,
        scenario_11a,
        scenario_11b,
        scenario_11c,
        scenario_11d,
        scenario_12_v1,
        scenario_12_v2,
        scenario_12_v3,
    ]


    def test_runner(self):
        # randomize the order of the test execution to detect hidden dependencies
        random.shuffle(self.param_scenario)
        for simulation_file, system_file, sensitivity_file, expected_number_file_result, expected_result_list in self.param_scenario:
            with self.subTest():
                self._runner_E2E(simulation_file,
                                 system_file,
                                 sensitivity_file,
                                 expected_number_file_result,
                                 expected_result_list)

    def _runner_E2E(self,
                    simulation_file,
                    system_file,
                    sensitivity_file,
                    expected_number_file_result,
                    expected_result_list):

        output_folder = "./availsim4core/test/E2E/output/phases/"

        # Clean folder
        if os.path.exists(output_folder):
            shutil.rmtree(output_folder)

        # Run the main process
        main.start(simulation_file, system_file, sensitivity_file, output_folder)

        result_simulation_file_list = glob.glob(output_folder + "/*.xlsx")
        self.assertEqual(len(result_simulation_file_list), expected_number_file_result)

        for expected_result in expected_result_list:
            actual_result = expected_result.extract_result(result_simulation_file_list)

            if expected_result.tolerance == -1:
                self.assertEqual(expected_result.expected_result, actual_result)
            else:
                self.assertAlmostEqual(expected_result.expected_result, actual_result, places=expected_result.tolerance)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from typing import List

from availsim4core.src.context.system.component_tree.status import Status
from availsim4core.src.context.system.probability_law.probability_law import ProbabilityLaw
from availsim4core.src.context.system.probability_law.probability_law_factory import ProbabilityLawFactory


class MinimalReplaceableUnit:
    """
    Domain object class to represent a MinimalReplaceableUnit.
    This class represent both the TRIGGER mru and TARGET mru.
    """

    def __init__(self,
                 name: str,
                 repair_law: ProbabilityLaw,
                 repair_schedule: str,
                 triggering_status: Status,
                 scope_common_ancestor: List[str]):
        self.name = name
        self.repair_law: ProbabilityLaw = repair_law
        self.repair_schedule: str = repair_schedule
        self.status = triggering_status
        self.scope_common_ancestor = scope_common_ancestor

    def __hash__(self):
        return hash((type(self), self.name,
                     self.repair_schedule,
                     self.status,
                     *self.scope_common_ancestor))

    def __str__(self):
        return f"{self.name} - {self.repair_law} - {self.repair_schedule} - {self.status} - {self.scope_common_ancestor}"

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        if not isinstance(other, MinimalReplaceableUnit):
            return NotImplemented
        return self.name == other.name and \
               self.repair_law == other.repair_law and \
               self.repair_schedule == other.repair_schedule and \
               self.status == other.status and \
               self.scope_common_ancestor == other.scope_common_ancestor


class MinimalReplaceableUnitFactory:
    """
    Utility class to build a 'MinimalReplaceableUnit'.
    """

    @staticmethod
    def build(mru_name,
              mru_repair_law_name,
              mru_repair_parameters,
              minimal_replaceable_repair_schedule,
              triggering_status_list_of_str,
              scope_common_ancestor_name_str) -> List[MinimalReplaceableUnit]:
        """
        This methods returns a List of {MinimalReplaceableUnit}.
        For each `triggering_status` it creates a different mru.
        """
        return [MinimalReplaceableUnit(mru_name,
                                       ProbabilityLawFactory.build(mru_repair_law_name,
                                                                   mru_repair_parameters),
                                       minimal_replaceable_repair_schedule,
                                       Status(triggering_status),
                                       scope_common_ancestor_name_str)
                for triggering_status in triggering_status_list_of_str]



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from availsim4core.src.reader.xlsx.monte_carlo_reader import MonteCarloReader
from availsim4core.src.reader.xlsx.monte_carlo_reader import SimulationMonteCarloColumn
from availsim4core.src.simulation.splitting_mc import SplittingMonteCarlo

class SimulationSplittingMonteCarloColumn:
    SYSTEM_ROOT_COMPONENT = "SYSTEM_ROOT_COMPONENT"


class SplittingMonteCarloReader(MonteCarloReader):
    """
    Specific reader for the MonteCarlo simulation.
    """

    @classmethod
    def build(cls, simulation_parameters):
        return SplittingMonteCarlo(
            simulation_parameters[SimulationMonteCarloColumn.MINIMUM_NUMBER_OF_SIMULATION],
            simulation_parameters[SimulationMonteCarloColumn.MAXIMUM_NUMBER_OF_SIMULATION],
            simulation_parameters[SimulationMonteCarloColumn.CONVERGENCE_MARGIN],
            simulation_parameters[SimulationMonteCarloColumn.MAXIMUM_EXECUTION_TIME],
            simulation_parameters[SimulationMonteCarloColumn.SEED],
            cls.extract_from_str_list(simulation_parameters[SimulationMonteCarloColumn.DIAGNOSTICS]),
            simulation_parameters[SimulationMonteCarloColumn.DURATION], 
            cls.clean_cell_str_else_raise(simulation_parameters[SimulationSplittingMonteCarloColumn.SYSTEM_ROOT_COMPONENT],
                                          exception_message_hint='simulation file'))


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest

from availsim4core.src.context.phase.phase import Phase
from availsim4core.src.context.phase.phase_manager import PhaseManager
from availsim4core.src.context.system.architecture_entry import ArchitectureEntry
from availsim4core.src.context.system.failure import Failure, FailureType
from availsim4core.src.context.system.failure_mode import FailureMode
from availsim4core.src.context.system.failure_mode_assignments import FailureModeAssignments
from availsim4core.src.context.system.inspection import Inspection
from availsim4core.src.context.system.minimal_replaceable_unit import MinimalReplaceableUnitFactory
from availsim4core.src.context.system.probability_law.probability_law import ProbabilityLaw
from availsim4core.src.reader.xlsx.system_template_reader import SystemTemplateReader, DuplicatedEntry


class test_SystemTemplateReader(unittest.TestCase):

    def test_generate_failure_mode_assignments(self):
        system_dictionary_assignments = {
            0: {"COMPONENT_NAME": 'ACCELERATORCONTROLS', "FAILURE_MODE_NAME": 'ACCELERATORCONTROLSFAILURE'},
            1: {"COMPONENT_NAME": 'ACCESSMANAGEMENT', "FAILURE_MODE_NAME": 'ACCESSMANAGEMENTFAILURE'}
        }

        failure_mode_1 = FailureMode("ACCELERATORCONTROLSFAILURE",
                                     ProbabilityLaw("", 0, False),
                                     ProbabilityLaw("", 0, False),
                                     Failure(FailureType.DETECTABLE),
                                     "[]",
                                     None,
                                     [],
                                     None,
                                     'AFTER_REPAIR',
                                     [])

        failure_mode_2 = FailureMode("ACCESSMANAGEMENTFAILURE",
                                     ProbabilityLaw("", 0, False),
                                     ProbabilityLaw("", 0, False),
                                     Failure(FailureType.DETECTABLE),
                                     "[]",
                                     None,
                                     [],
                                     None,
                                     'AFTER_REPAIR',
                                     [])

        failure_mode_assignment_1 = FailureModeAssignments("ACCELERATORCONTROLS", failure_mode_1)
        failure_mode_assignment_2 = FailureModeAssignments("ACCESSMANAGEMENT", failure_mode_2)

        expected_results = [failure_mode_assignment_1, failure_mode_assignment_2]

        results = SystemTemplateReader.generate_failure_mode_assignments(system_dictionary_assignments,
                                                                         [failure_mode_1, failure_mode_2])

        self.assertEqual(expected_results, results)

    def test_generate_failure_modes(self):
        failure_mode_dict = {
            0: {"FAILURE_MODE_NAME": 'ACCELERATORCONTROLSFAILURE',
                "FAILURE_LAW": "EXP",
                "FAILURE_PARAMETERS": 300.0,
                "REPAIR_LAW": "FIX",
                "REPAIR_PARAMETERS": 1.5436,
                "TYPE_OF_FAILURE": "DETECTABLE",
                "HELD_BEFORE_REPAIR": "HELD_FOREVER",
                "INSPECTION_NAME": 'none',
                "PHASE_NAME": 'A',
                "NEXT_PHASE_IF_FAILURE": 'B',
                "PHASE_CHANGE_TRIGGER": "AFTER_FAILURE",
                "HELD_AFTER_REPAIR":"HELD_FOREVER"},
            1: {"FAILURE_MODE_NAME": 'ACCESSMANAGEMENTFAILURE',
                "FAILURE_LAW": "EXP",
                "FAILURE_PARAMETERS": 859.74,
                "TYPE_OF_FAILURE": "DETECTABLE",
                "HELD_BEFORE_REPAIR": "HELD_FOREVER",
                "REPAIR_LAW": "FIX",
                "REPAIR_PARAMETERS": 3.6167,
                "INSPECTION_NAME": 'none',
                "PHASE_NAME": 'B',
                "NEXT_PHASE_IF_FAILURE": 'none',
                "PHASE_CHANGE_TRIGGER": "AFTER_FAILURE",
                "HELD_AFTER_REPAIR":"HELD_FOREVER"},
        }

        phase_a = Phase('A', ProbabilityLaw("FIX", 1., False), True)
        phase_b = Phase('B', ProbabilityLaw("FIX", 1., False), False)
        phase_a.set_next_phase(phase_b)
        phase_b.set_next_phase(phase_a)

        failure_mode_1 = FailureMode("ACCELERATORCONTROLSFAILURE",
                                     ProbabilityLaw("ExponentialLaw", [300.0], False),
                                     ProbabilityLaw("DeterministicLaw", [1.5436], False),
                                     Failure(FailureType.DETECTABLE),
                                     set([PhaseManager.HELD_FOREVER]),
                                     None,
                                     {phase_a},
                                     phase_b,
                                     "AFTER_FAILURE",
                                     set([PhaseManager.HELD_FOREVER]))
        failure_mode_2 = FailureMode("ACCESSMANAGEMENTFAILURE",
                                     ProbabilityLaw("ExponentialLaw", [859.74], False),
                                     ProbabilityLaw("DeterministicLaw", [3.6167], False),
                                     Failure(FailureType.DETECTABLE),
                                     set([PhaseManager.HELD_FOREVER]),
                                     None,
                                     {phase_b},
                                     None,
                                     "AFTER_FAILURE",
                                     set([PhaseManager.HELD_FOREVER]))

        expected_results = [failure_mode_1, failure_mode_2]

        results = SystemTemplateReader.generate_failure_modes(failure_mode_dict, {}, {phase_a, phase_b})

        self.assertEqual(expected_results, results)

    def test_generate_architecture_entry_list(self):
        architecture_dict = {
            0: {"COMPONENT_NAME": 'ROOT',
                "COMPONENT_TYPE": 'COMPOUND',
                "COMPONENT_NUMBER": 1,
                "CHILDREN_NAME": 'ACCELERATORCONTROLS,ACCESSMANAGEMENT,BEAMLOSSES',
                "CHILDREN_LOGIC": "AND",
                "IN_MRU": 'none',
                "TRIGGER_MRU": 'none'},
            1: {"COMPONENT_NAME": 'ACCELERATORCONTROLS',
                "COMPONENT_TYPE": 'basic',
                "COMPONENT_NUMBER": 2,
                "CHILDREN_NAME": 'none',
                "CHILDREN_LOGIC": 'none',
                "IN_MRU": 'none',
                "TRIGGER_MRU": 'none'}
        }

        expected_architecture_entry_root = ArchitectureEntry('ROOT', "COMPOUND", 1,
                                                             ['ACCELERATORCONTROLS', 'ACCESSMANAGEMENT', 'BEAMLOSSES'],
                                                             "AND",
                                                             [], [])
        expected_architecture_entry_basic = ArchitectureEntry('ACCELERATORCONTROLS', "BASIC", 2, [], '', [], [])

        architecture_list_result = SystemTemplateReader.generate_architecture_entry_list(architecture_dict)

        self.assertEqual(architecture_list_result,
                         [expected_architecture_entry_root, expected_architecture_entry_basic])

    def test_generate_architecture_entry_list_with_duplicated_entries(self):
        architecture_dict = {
            0: {"COMPONENT_NAME": 'ROOT',
                "COMPONENT_TYPE": 'COMPOUND',
                "COMPONENT_NUMBER": 1,
                "CHILDREN_NAME": 'ACCELERATORCONTROLS,ACCESSMANAGEMENT,BEAMLOSSES',
                "CHILDREN_LOGIC": "AND",
                "IN_MRU": 'none',
                "TRIGGER_MRU": 'none'},
            1: {"COMPONENT_NAME": 'ACCELERATORCONTROLS',
                "COMPONENT_TYPE": 'basic',
                "COMPONENT_NUMBER": 2,
                "CHILDREN_NAME": 'none',
                "CHILDREN_LOGIC": 'none',
                "IN_MRU": 'none',
                "TRIGGER_MRU": 'none'},
            2: {"COMPONENT_NAME": 'ACCELERATORCONTROLS',
                "COMPONENT_TYPE": 'basic',
                "COMPONENT_NUMBER": 2,
                "CHILDREN_NAME": 'none',
                "CHILDREN_LOGIC": 'none',
                "IN_MRU": 'none',
                "TRIGGER_MRU": 'none'}
        }

        with self.assertRaises(DuplicatedEntry) as context:
            architecture_list_result = SystemTemplateReader.generate_architecture_entry_list(architecture_dict)

        self.assertTrue("The component name ACCELERATORCONTROLS is not unique in the list of architecture entries" in str(context.exception))

    def test_generate_minimal_replaceable_units(self):
        system_dictionary_minimal_replaceable_unit = {
            0: {'MRU_NAME': 'CLIQ',
                "MRU_LAW": "FIX",
                "MRU_PARAMETERS": 0.001,
                "MRU_SCHEDULE": "IMMEDIATE",
                "LOWEST_COMMON_ANCESTOR_SCOPE": 'CLIQ',
                "TRIGGERING_STATUS": 'FAILED'},
            1: {'MRU_NAME': 'SHUTDOWN',
                "MRU_LAW": "FIX",
                "MRU_PARAMETERS": 0.001,
                "MRU_SCHEDULE": "IMMEDIATE",
                "LOWEST_COMMON_ANCESTOR_SCOPE": 'ITS',
                "TRIGGERING_STATUS": 'UNDER_REPAIR'},
            2: {'MRU_NAME': 'REPAIRALL',
                "MRU_LAW": "FIX",
                "MRU_PARAMETERS": 0.675,
                "MRU_SCHEDULE": "IMMEDIATE",
                "LOWEST_COMMON_ANCESTOR_SCOPE": 'IT',
                "TRIGGERING_STATUS": 'FAILED,BLIND_FAILED'}
        }

        result = SystemTemplateReader.generate_mrus(system_dictionary_minimal_replaceable_unit)

        expected_result = \
            MinimalReplaceableUnitFactory.build("CLIQ",
                                                "FIX",
                                                [0.001],
                                                "IMMEDIATE",
                                                ["FAILED"],
                                                ["CLIQ"]) + \
            MinimalReplaceableUnitFactory.build("SHUTDOWN",
                                                "FIX",
                                                [0.001],
                                                "IMMEDIATE",
                                                ["UNDER_REPAIR"],
                                                ["ITS"]) + \
            MinimalReplaceableUnitFactory.build('REPAIRALL',
                                                "FIX",
                                                [0.675],
                                                "IMMEDIATE",
                                                ['FAILED', 'BLIND_FAILED'],
                                                ["IT"])

        self.assertListEqual(result, expected_result)

    def test_generate_minimal_replaceable_units_with_multiple_lca(self):
        system_dictionary_minimal_replaceable_unit = {
            0: {'MRU_NAME': 'CLIQ',
                "MRU_LAW": "FIX",
                "MRU_PARAMETERS": 0.001,
                "MRU_SCHEDULE": "IMMEDIATE",
                "LOWEST_COMMON_ANCESTOR_SCOPE": '[CLIQ,QH]',
                "TRIGGERING_STATUS": 'FAILED'},
            1: {'MRU_NAME': 'SHUTDOWN',
                "MRU_LAW": "FIX",
                "MRU_PARAMETERS": 0.001,
                "MRU_SCHEDULE": "IMMEDIATE",
                "LOWEST_COMMON_ANCESTOR_SCOPE": 'ITS',
                "TRIGGERING_STATUS": 'UNDER_REPAIR'},
            2: {'MRU_NAME': 'REPAIRALL',
                "MRU_LAW": "FIX",
                "MRU_PARAMETERS": 0.675,
                "MRU_SCHEDULE": "IMMEDIATE",
                "LOWEST_COMMON_ANCESTOR_SCOPE": '[IT1,IT2]',
                "TRIGGERING_STATUS": 'FAILED,BLIND_FAILED'}
        }

        result = SystemTemplateReader.generate_mrus(system_dictionary_minimal_replaceable_unit)

        expected_result = \
            MinimalReplaceableUnitFactory.build("CLIQ",
                                                "FIX",
                                                [0.001],
                                                "IMMEDIATE",
                                                ["FAILED"],
                                                ["CLIQ","QH"]) + \
            MinimalReplaceableUnitFactory.build("SHUTDOWN",
                                                "FIX",
                                                [0.001],
                                                "IMMEDIATE",
                                                ["UNDER_REPAIR"],
                                                ["ITS"]) + \
            MinimalReplaceableUnitFactory.build('REPAIRALL',
                                                "FIX",
                                                [0.675],
                                                "IMMEDIATE",
                                                ['FAILED', 'BLIND_FAILED'],
                                                ["IT1","IT2"])

        self.assertListEqual(result, expected_result)

    def test_generate_inspections(self):
        system_dictionary_inspections = \
            {
                0: {"INSPECTION_NAME": 'ITINSPECTION',
                    "INSPECTION_PERIOD": 4.999,
                    "INSPECTION_DURATION": 0.001}
            }

        result = SystemTemplateReader.generate_inspections(system_dictionary_inspections)

        expected_result = [
            Inspection("ITINSPECTION",
                       4.999,
                       0.001)
        ]

        self.assertListEqual(result, expected_result)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import shutil

import pandas

from availsim4core.src.analysis import Analysis
from availsim4core.src.context.context import Context
from availsim4core.src.context.system.component_tree.status import Status
from availsim4core.src.exporter.result_exporter import ResultExporter
from availsim4core.src.exporter.xlsx_analysis_exporter import XLSXAnalysisExporter
from availsim4core.src.results.results import Results


class XLSXResultExporterSummary(ResultExporter):
    """
    class used to export the results of a simulation into an output file
    This exporter is meant to provide a good "summary" of a simulation, not a light report like the exporter LightStudy
    """

    @classmethod
    def parse_statistics(cls,
                         statistics,
                         type_of_statistics_str: str):
        """
        Function extracting the statistics into a list of dictionaries
        """

        result_list = []

        for result_record_entry, statistics in statistics.items():
            result_record_entry_attributes = dir(result_record_entry)
            component = result_record_entry.identifier()
            phase = result_record_entry.identifier() if "phase" not in result_record_entry_attributes \
                else getattr(result_record_entry, 'phase').name
            status = "_" if "status" not in result_record_entry_attributes \
                else getattr(result_record_entry, 'status')
            description = "_" if "description" not in result_record_entry_attributes \
                else getattr(result_record_entry, 'description')
            occurrences_key_value = {
                "component": component,
                "phase": phase,
                "status": status,
                "description": description,
                # TODO availsim4/availsim4core#62 BUG in the way _MIN is computed
                # "_MIN_" + type_of_statistics_str: statistics.min,
                # the value 0 cannot appear in _MIN_ because we only count what exists
                # one solution is to generate every possible combination of record and check that they appear 0 time.
                "_MEAN_" + type_of_statistics_str: statistics.mean,
                "_MAX_" + type_of_statistics_str: statistics.max,
                "_STD_" + type_of_statistics_str: statistics.std,
                "_BOOLEAN_" + type_of_statistics_str: statistics.boolean
            }
            result_list.append(occurrences_key_value)

        return result_list

    @classmethod
    def root_statistics(cls,
                        result_dataframe,
                        root_identifier: str,
                        writer):

        root_result_dataframe = result_dataframe.loc[result_dataframe['component'] == root_identifier]

        root_summary_result = {
            "uptime_duration": 0.0,
            "downtime_duration": 0.0,
        }

        list_of_uptime_status = [Status.RUNNING, Status.DEGRADED, Status.BLIND_DEGRADED]
        for status in list_of_uptime_status:
            root_summary_result["uptime_duration"] += root_result_dataframe.loc[
                root_result_dataframe["status"] == status, "_MEAN_DURATION"
            ].sum()

        list_of_downtime_status = [Status.FAILED, Status.BLIND_FAILED, Status.INSPECTION, Status.UNDER_REPAIR]
        for status in list_of_downtime_status:
            root_summary_result["downtime_duration"] += root_result_dataframe.loc[
                root_result_dataframe["status"] == status, "_MEAN_DURATION"
            ].sum()

        root_summary_result["total"] = root_summary_result["uptime_duration"] + root_summary_result["downtime_duration"]
        root_summary_result["availability"] = root_summary_result["uptime_duration"] / root_summary_result["total"]
        root_summary_result_dataframe = pandas.DataFrame(root_summary_result, index=[0])

        root_summary_result_dataframe.to_excel(writer, sheet_name='RESULTS_ROOT_SUMMARY', index=False)

    @classmethod
    def phase_statistics(cls,
                         result_dataframe,
                         root_identifier: str,
                         writer):

        list_of_phases_start_ok = ["default", "Init"]
        list_of_phases_start_not_ok = ["default failure", "specific failure"]
        list_of_phases_start_specific = ["specific failure"]

        simulation_duration = result_dataframe.loc[
            (result_dataframe['component'] == 'Phase'),
            "_MEAN_DURATION"].sum()

        result_dataframe['start_of_phase_ok'] = result_dataframe.loc[
            (result_dataframe['component'] == 'Phase') &
            (result_dataframe["description"].isin(list_of_phases_start_ok)),
            "_MEAN_OCCURRENCES"].fillna(0)

        result_dataframe['start_of_phase_not_ok'] = result_dataframe.loc[
            (result_dataframe['component'] == 'Phase') &
            (result_dataframe["description"].isin(list_of_phases_start_not_ok)),
            "_MEAN_OCCURRENCES"].fillna(0)

        result_dataframe['start_of_phase_specific'] = result_dataframe.loc[
            (result_dataframe['component'] == 'Phase') &
            (result_dataframe["description"].isin(list_of_phases_start_specific)),
            "_MEAN_OCCURRENCES"].fillna(0)

        result_dataframe['failures_in_phase'] = result_dataframe.loc[
            (result_dataframe["component"] == root_identifier) &
            (result_dataframe["status"] == Status.FAILED),
            "_MEAN_OCCURRENCES"].fillna(0)

        result_dataframe['blind_failures_in_phase'] = result_dataframe.loc[
            (result_dataframe["component"] == root_identifier) &
            (result_dataframe["status"] == Status.BLIND_FAILED),
            "_MEAN_OCCURRENCES"].fillna(0)

        result_dataframe['RUNNING_in_phase'] = result_dataframe.loc[
            (result_dataframe["component"] == root_identifier) &
            (result_dataframe["status"] == Status.RUNNING),
            "_MEAN_DURATION"].fillna(0)

        result_dataframe['UNDER_REPAIR_in_phase'] = result_dataframe.loc[
            (result_dataframe["component"] == root_identifier) &
            (result_dataframe["status"] == Status.UNDER_REPAIR),
            "_MEAN_DURATION"].fillna(0)

        result_dataframe['DEGRADED_in_phase'] = result_dataframe.loc[
            (result_dataframe["component"] == root_identifier) &
            (result_dataframe["status"] == Status.DEGRADED),
            "_MEAN_DURATION"].fillna(0)

        result_dataframe['BLIND_DEGRADED_in_phase'] = result_dataframe.loc[
            (result_dataframe["component"] == root_identifier) &
            (result_dataframe["status"] == Status.BLIND_DEGRADED),
            "_MEAN_DURATION"].fillna(0)

        result_dataframe['FAILED_in_phase'] = result_dataframe.loc[
            (result_dataframe["component"] == root_identifier) &
            (result_dataframe["status"] == Status.FAILED),
            "_MEAN_DURATION"].fillna(0)

        result_dataframe['BLIND_FAILED_in_phase'] = result_dataframe.loc[
            (result_dataframe["component"] == root_identifier) &
            (result_dataframe["status"] == Status.BLIND_FAILED),
            "_MEAN_DURATION"].fillna(0)

        result_dataframe['INSPECTION_in_phase'] = result_dataframe.loc[
            (result_dataframe["component"] == root_identifier) &
            (result_dataframe["status"] == Status.INSPECTION),
            "_MEAN_DURATION"].fillna(0)

        result_dataframe = result_dataframe.set_index('phase')
        phases_data_gby = result_dataframe.loc[result_dataframe['component'] == 'Phase'].groupby('phase')

        summary_phases = phases_data_gby['_MEAN_OCCURRENCES'].sum().to_frame('OCCURRENCES')
        summary_phases['TOTAL_DURATION'] = phases_data_gby['_MEAN_DURATION'].sum()
        summary_phases['MEAN_DURATION'] = summary_phases['TOTAL_DURATION'] / summary_phases['OCCURRENCES']

        summary_phases['NORMAL_START_OF_A_PHASE'] = phases_data_gby['start_of_phase_ok'].sum()
        summary_phases['START_OF_A_PHASE_DUE_TO_A_FAILURE'] = phases_data_gby['start_of_phase_not_ok'].sum()
        summary_phases['START_OF_A_PHASE_DUE_TO_A_SPECIFIC_FAILURE'] = phases_data_gby['start_of_phase_specific'].sum()

        summary_phases['SUM_OF_DETECTABLE_FAILURES_IN_A_PHASE'] = result_dataframe['failures_in_phase'].groupby(
            'phase').sum()
        summary_phases['SUM_OF_BLIND_FAILURES_IN_A_PHASE'] = result_dataframe['blind_failures_in_phase'].groupby(
            'phase').sum()

        summary_phases['TOTAL_DURATION_OF_THE_RUNNING_STATUS'] = result_dataframe['RUNNING_in_phase'].groupby(
            'phase').sum()
        summary_phases['TOTAL_DURATION_OF_THE_UNDER_REPAIR_STATUS'] = result_dataframe['UNDER_REPAIR_in_phase'].groupby(
            'phase').sum()
        summary_phases['TOTAL_DURATION_OF_THE_DEGRADED_STATUS'] = result_dataframe['DEGRADED_in_phase'].groupby(
            'phase').sum()
        summary_phases['TOTAL_DURATION_OF_THE_BLIND_DEGRADED_STATUS'] = result_dataframe[
            'BLIND_DEGRADED_in_phase'].groupby(
            'phase').sum()
        summary_phases['TOTAL_DURATION_OF_THE_FAILED_STATUS'] = result_dataframe['FAILED_in_phase'].groupby(
            'phase').sum()
        summary_phases['TOTAL_DURATION_OF_THE_BLIND_FAILED_STATUS'] = result_dataframe['BLIND_FAILED_in_phase'].groupby(
            'phase').sum()
        summary_phases['TOTAL_DURATION_OF_THE_INSPECTION_STATUS'] = result_dataframe['INSPECTION_in_phase'].groupby(
            'phase').sum()

        summary_phases['RATIO_OF_PHASE_WITHOUT_DETECTABLE_FAILURES'] = (summary_phases['OCCURRENCES'] - summary_phases[
            'SUM_OF_DETECTABLE_FAILURES_IN_A_PHASE']) / summary_phases['OCCURRENCES']
        summary_phases['FRACTION_OF_UPTIME_WITHIN_THE_PHASE'] = (summary_phases[
                                                                     'TOTAL_DURATION_OF_THE_RUNNING_STATUS'] +
                                                                 summary_phases[
                                                                     'TOTAL_DURATION_OF_THE_DEGRADED_STATUS'] +
                                                                 summary_phases[
                                                                     'TOTAL_DURATION_OF_THE_BLIND_DEGRADED_STATUS']) / \
                                                                summary_phases['TOTAL_DURATION']

        summary_phases['FRACTION_OF_UPTIME_WITHIN_THE_TOTAL_DURATION'] = (
                                                                                 summary_phases[
                                                                                     'TOTAL_DURATION_OF_THE_RUNNING_STATUS'] +
                                                                                 summary_phases[
                                                                                     'TOTAL_DURATION_OF_THE_DEGRADED_STATUS'] +
                                                                                 summary_phases[
                                                                                     'TOTAL_DURATION_OF_THE_BLIND_DEGRADED_STATUS']) / simulation_duration

        summary_phases = summary_phases.reset_index()

        tmp = pandas.DataFrame(summary_phases.sum()).T
        tmp['phase'] = '__SUM_OVER_EVERY_PHASE__'
        summary_phases = summary_phases.append(tmp)

        summary_phases = summary_phases.rename(columns={'phase': 'PHASE'})

        summary_phases.to_excel(writer, sheet_name='RESULTS_PHASE_SUMMARY', index=False)

    @classmethod
    def export(cls,
               context: Context,
               output_folder: str,
               analysis: Analysis,
               results: Results,
               execution_time: float,
               unique_output_file_identifier_str: str):

        output_file = cls.OUTPUT_FILE_TEMPLATE.format(output_folder, analysis.id, unique_output_file_identifier_str)
        writer = ResultExporter.open_workbook(output_file)

        XLSXAnalysisExporter.export(writer,
                                    analysis)

        result_list_occurrence = cls.parse_statistics(results.occurrence_statistics, "OCCURRENCES")
        result_list_duration = cls.parse_statistics(results.duration_statistics, "DURATION")

        # save the result
        result_dataframe_occurrence = pandas.DataFrame(result_list_occurrence)
        result_dataframe_duration = pandas.DataFrame(result_list_duration)

        result_dataframe = pandas.merge(result_dataframe_occurrence, result_dataframe_duration,
                                        on=["component", "phase", "status", "description"])

        # computing some simple and useful stat
        root_component = context.root_component
        root_identifier = f"{root_component.name}_{root_component.local_id}_{root_component.global_id}"
        # computing some simple and useful stat of the root component
        cls.root_statistics(result_dataframe.copy(), root_identifier, writer)
        # computing some simple and useful stat on the phases
        cls.phase_statistics(result_dataframe.copy(), root_identifier, writer)

        result_dataframe.to_excel(writer, sheet_name='RESULTS', index=False)

        writer.save()

        # export
        if analysis.system_template.custom_children_logic_path:
            shutil.copy(analysis.system_template.custom_children_logic_path,
                        output_folder+analysis.system_template.custom_children_logic_path.rsplit('/',1)[-1])


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from collections import Counter, defaultdict

from availsim4core.src.results.results import Results
from availsim4core.src.statistics.statistics import Statistics
from availsim4core.src.timeline.timeline import Timeline


class SimulationResults(Results):
    """
    statistics used to profile the execution of the code
    """

    def __init__(self,
                 maximum_number_of_simulations,
                 duration_of_a_simulation):
        self.maximum_number_of_simulations = maximum_number_of_simulations
        self.duration_of_a_simulation = duration_of_a_simulation

        self.number_of_DES_simulations_executed = {}

        self.execution_metrics_statistics = {}
        self.occurrence_statistics = {}
        self.duration_statistics = {}

        self.last_simulation_timeline = []
        self.root_cause_analysis_record_list = []

    def __str__(self):
        return str(self.__dict__)

    def update_with_des_results(self,
                                execution_metrics: dict,
                                timeline: Timeline, 
                                weight: float = 1):
        """
        Update the results given the metric about the code execution and the full timeline of the discrete event simulation.
        :return:
        """
        self._update_execution_metrics(execution_metrics, weight)
        self._update_occurrence_statistics(timeline, weight)
        self._update_duration_statistics(timeline, weight)

    def _update_execution_metrics(self, execution_metrics, weight):

        for key, value in execution_metrics.items():
            if key not in self.execution_metrics_statistics.keys():
                self.execution_metrics_statistics[key] = Statistics()
            statistics = self.execution_metrics_statistics[key]
            statistics.update(value, weight)

    def _update_occurrence_statistics(self, timeline, weight):

        result_record_entry_list = list(map(lambda record: record.get_result_record_entry(),
                                            timeline.record_list))
        counter = Counter(result_record_entry_list)

        for result_record_entry, occurrence in counter.items():
            if result_record_entry not in self.occurrence_statistics:
                self.occurrence_statistics[result_record_entry] = Statistics()
            statistics = self.occurrence_statistics[result_record_entry]
            statistics.update(occurrence, weight)

    def _update_duration_statistics(self, timeline, weight):
        def full_group_by(l, key=lambda x: x):
            d = defaultdict(list)
            for item in l:
                d[key(item)].append(item)
            return d.values()

        record_list_sorted_by_identifier = [g
                                            for g in full_group_by(timeline.record_list,
                                                                   key=lambda
                                                                       r: r.get_result_record_entry().identifier())]

        result_record_entry_duration_dict = {}
        for record_list in record_list_sorted_by_identifier:
            sorted_record_list = record_list
            for index, record in enumerate(sorted_record_list):
                if index + 1 < len(sorted_record_list):
                    duration = sorted_record_list[index + 1].timestamp - record.timestamp
                else:
                    duration = self.duration_of_a_simulation - record.timestamp

                result_record_entry = record.get_result_record_entry()
                if result_record_entry in result_record_entry_duration_dict:
                    result_record_entry_duration_dict[result_record_entry] += duration
                else:
                    result_record_entry_duration_dict[result_record_entry] = duration

        for result_record_entry, record_duration in result_record_entry_duration_dict.items():
            if result_record_entry not in self.duration_statistics:
                self.duration_statistics[result_record_entry] = Statistics()
            statistics = self.duration_statistics[result_record_entry]
            statistics.update(record_duration, weight)

    def evaluate_result(self):

        for statistics in self.execution_metrics_statistics.values():
            statistics.evaluate_result()

        for statistics in self.occurrence_statistics.values():
            statistics.evaluate_result()

        for statistics in self.duration_statistics.values():
            statistics.evaluate_result()



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import pandas

from availsim4core.src.analysis import Analysis
from availsim4core.src.context.context import Context
from availsim4core.src.exporter.result_exporter import ResultExporter
from availsim4core.src.results.results import Results


class XLSXResultExporterRCA(ResultExporter):
    @classmethod
    def export(cls,
               context: Context,
               output_folder: str,
               analysis: Analysis,
               results: Results,
               execution_time: float,
               unique_output_file_identifier_str: str):

        output_file = cls.OUTPUT_FILE_TEMPLATE.format(output_folder, analysis.id, unique_output_file_identifier_str)
        writer = ResultExporter.open_workbook(output_file)

        rca_export = []
        for record in results.root_cause_analysis_record_list:
            rca_export.append({
                "analysis_id": record.simulation_id,
                "timestamp": record.timestamp, 
                "rca_component_trigger": record.rca_trigger_component,
                "rca_status_trigger": record.rca_trigger_status,
                "rca_phase_trigger": record.rca_trigger_phase,
                "rca_event_description": record.description,
                **record.component_statuses_dict
            })
        rca_export_dataframe = pandas.DataFrame(rca_export)
        rca_export_dataframe.to_excel(writer, sheet_name="RESULTS_ROOT_CAUSE_ANALYSIS", index=False)

        writer.save()


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from availsim4core.src.context.context import Context
from availsim4core.src.discrete_event_simulation.event.c_event.c_event import CEvent


class OrderNextPhaseEvent(CEvent):
    """
    CEvent class in charge of creating a BEvent {NextPhaseEvent}.
    """

    def __init__(self,
                 priority: int,
                 context: Context):
        super().__init__(priority, context)

    def __eq__(self, other):
        if not isinstance(self, OrderNextPhaseEvent):
            return NotImplemented
        return super().__eq__(other)

    def __hash__(self):
        return hash((type(self), self.priority))

    def __str__(self):
        return f"OrderNextPhaseEvent:: " \
               f"priority:{self.priority}"

    def generate_b_events(self, absolute_simulation_time):
        from availsim4core.src.discrete_event_simulation.event.b_event.phase_event.next_phase_event import \
            NextPhaseEvent
        next_phase_time = absolute_simulation_time + self.context.phase_manager.current_phase.law.get_random_value()
        event = NextPhaseEvent(next_phase_time,
                               self.context)
        return {event}



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import pandas

from openpyxl import load_workbook
from availsim4core.src.analysis import Analysis
from availsim4core.src.context.context import Context
from availsim4core.src.results.results import Results


class ResultExporter:
    """
    Interface to define the exporter api for each .
    """

    OUTPUT_FILE_TEMPLATE = '{}simulation_result_{}_{}.xlsx'

    @classmethod
    def export(cls,
               context: Context,
               output_folder: str,
               analysis: Analysis,
               results: Results,
               execution_time: float,
               unique_output_file_identifier_str):
        f"""
        Exports results.
        :param context {Context} of the result to export.
        :param output_folder folder to export result.
        :param analysis {Analysis} the analysis which have been proceed.
        :param results {Results} the simulation results.
        :param execution_time time of the simulation.
        :param unique_output_file_identifier_str to uniquely identify group of output files. Currently the date/time at exporting is used as a label.
        """
        pass

    @classmethod
    def open_workbook(cls,
                      output_file):
        """
        Function checking if the output file already exists and returning a writer
        :param output_file: name of the output file
        :return writer: object allowing to write sheets into an excel file using dataframe as the source of data
        """

        try:
            preexisting_data = load_workbook(output_file)
        except FileNotFoundError:
            preexisting_data = None

        writer = pandas.ExcelWriter(output_file, engine='openpyxl')

        if preexisting_data is not None:
            writer.book = preexisting_data

        return writer


# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 


import copy
import logging
from datetime import datetime
from typing import List

from availsim4core.src.context.context import Context
from availsim4core.src.discrete_event_simulation.discrete_event_simulation import DiscreteEventSimulation
from availsim4core.src.results.simulation_results import SimulationResults
from availsim4core.src.simulation.simulation import Simulation, SimulationType


class MonteCarlo(Simulation):
    """
    Defines a MonteCarlo simulation that will be triggered by the Sensitivity Analysis.
    Attributes:
    - minimum_number_of_simulations: minimun number of call to discrete event simulation.
    - maximum_number_of_simulations: maximun number of call to discrete event simulation.
    - convergence_margin: Evaluate on the fly the appropriate number of simulation to run.
    - maximum_execution_time: used on the Cluster to stop the simulation after this period of time.
    - seed: Control the random generator to get different values over each simulation. Deterministic approach.
    - list_of_diagnosis: Defines the metrics to compute.
    - result_simulation: The ResultSimulation to return.
    Extends the simulation, by implementing the run() method.
    """
    SIMULATION_TYPE = SimulationType.MONTE_CARLO

    def __init__(self,
                 minimum_number_of_simulations: int,
                 maximum_number_of_simulations: int,
                 convergence_margin: float,
                 maximum_execution_time: float,
                 seed: int,
                 list_of_diagnosis: List[str],
                 duration_of_simulation: float):
        self.minimum_number_of_simulations = minimum_number_of_simulations
        self.maximum_number_of_simulations = maximum_number_of_simulations
        self.convergence_margin = convergence_margin
        self.maximum_execution_time = maximum_execution_time
        self.seed = seed
        self.list_of_diagnosis = list_of_diagnosis
        self.duration_of_simulation = duration_of_simulation

    def __str__(self):
        return f"{self.SIMULATION_TYPE}:: " \
               f"minimum_number_of_simulations: {self.minimum_number_of_simulations} - " \
               f"maximum_number_of_simulations: {self.maximum_number_of_simulations} - " \
               f"maximum_execution_time: {self.maximum_execution_time} - " \
               f"diagnostics: {self.list_of_diagnosis} - " \
               f"seed: {self.seed} - " \
               f"duration_of_simulation: {self.duration_of_simulation}"

    def __eq__(self, other):
        return self.minimum_number_of_simulations == other.minimum_number_of_simulations \
               and self.maximum_number_of_simulations == other.maximum_number_of_simulations \
               and self.convergence_margin == other.convergence_margin \
               and self.maximum_execution_time == other.maximum_execution_time \
               and self.seed == other.seed \
               and self.list_of_diagnosis == other.list_of_diagnosis \
               and self.duration_of_simulation == other.duration_of_simulation


    def update_statistics(self, simulation_results: SimulationResults, simulation_des, start_time, simulation_weight: float, recursion_depth):
        """
        Function called after each iteration of the Monte Carlo algorithm to update the statistics incrementally
        :param simulation_results:
        :param simulation_des:
        :param start_time:
        :param simulation_index:
        :return:
        """
        simulation_results.update_with_des_results(simulation_des.execution_metrics,
                                        simulation_des.context.timeline_record,
                                        simulation_weight)
        execution_time = (datetime.now() - start_time).total_seconds()
        # logging.info(f"DES simulation completed in: {execution_time:.2f} s; ")
        simulation_results.root_cause_analysis_record_list.extend(simulation_des.context.rca_manager.root_cause_analysis_record_list)
        if recursion_depth not in simulation_results.number_of_DES_simulations_executed:
            simulation_results.number_of_DES_simulations_executed[recursion_depth] = 0
        simulation_results.number_of_DES_simulations_executed[recursion_depth] += 1

    def run(self, context: Context):
        simulation_results = SimulationResults(self.maximum_number_of_simulations,
                                    self.duration_of_simulation)
        simulation_weight = 1/self.maximum_number_of_simulations

        for simulation_index in range(0, self.maximum_number_of_simulations):
            start_time = datetime.now()
            logging.info(f"Starting Monte Carlo simulation {simulation_index+1} / {self.maximum_number_of_simulations}")

            local_context = copy.deepcopy(context)
            # initialisation of a DES
            simulation_des = DiscreteEventSimulation(self.seed + simulation_index,
                                                     self.duration_of_simulation,
                                                     local_context, None, -1 , 0)

            # simulation
            _, _ = simulation_des.run()
            timeline_record = local_context.timeline_record

            # updating statistics
            self.update_statistics(simulation_results,
                                   simulation_des,
                                   start_time,
                                   simulation_weight,
                                   0)


        simulation_results.evaluate_result()
        simulation_results.last_simulation_timeline = timeline_record.record_list
        return simulation_results

    def get_list_of_diagnosis(self):
        return self.list_of_diagnosis



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import itertools

from availsim4core.src.sensitivity_analysis.exploration_strategy import Outer
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier import SystemModifier
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier_combination import SystemModifierCombination
from availsim4core.src.sensitivity_analysis.system_modifier.system_modifier_combination_strategy import \
    SystemModifierCombinationStrategy


class SystemModifierCombinationOuterStrategy(SystemModifierCombinationStrategy):
    """
       Strategy to apply the Outer logic.

    :eg.
       Default analysis with
           - A:0, B:0, C:0
       Given:
           - Outer(A, 1)
           - Outer(A, 2)
           - Outer(C, 3)
           - Outer(C, 4)

       It results in 4 analysis :
           - A:1, B:0, C:3
           - A:2, B:0, C:3
           - A:1, B:0, C:4
           - A:2, B:0, C:4
    """

    @classmethod
    def execute(cls, exploration_strategy_list):
        """
        Given a list of {ExplorationStrategy} it returns a list of corresponding {SystemModifierCombination}.
        :param exploration_strategy_list the list from which the Inner strategy will be extracted and applied to get
        the corresponding combination
        :returns list of {SystemModifierCombination} corresponding to the Outer logic.
        """
        outer_exploration_strategy_list = cls.filter_exploration(exploration_strategy_list,
                                                                 Outer)
        outer_combination_list = []
        for values_tuple in itertools.product(*outer_exploration_strategy_list):  # Cartesian product
            system_modifier_list = []
            for idx, value in enumerate(values_tuple):
                system_modifier_list.append(
                    SystemModifier(outer_exploration_strategy_list[idx].parameter_name, value))
            if system_modifier_list:
                outer_combination_list.append(SystemModifierCombination(system_modifier_list))
        return outer_combination_list



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from availsim4core.src.context.system.component_tree.basic import Component

from availsim4core.src.context.context import Context
from availsim4core.src.discrete_event_simulation.event.c_event.c_event import CEvent


class ComponentCEvent(CEvent):
    __slots__ = 'component'

    def __init__(self,
                 priority: int,
                 context: Context,
                 component: Component):
        super().__init__(priority, context)
        self.component = component

    def __eq__(self, other):
        if not isinstance(self, ComponentCEvent):
            return NotImplemented
        return super().__eq__(other) and \
            self.component == other.component

    def __lt__(self, other):
        if not isinstance(self, ComponentCEvent):
            return super().__lt__(self, ComponentCEvent)
        else:
            return super().__lt__(other) or (super().__eq__(other) and self.component.global_id < other.component.global_id)

    def __hash__(self):
        return hash((type(self), self.priority, self.component))

    def __str__(self):
        return (f"failure ComponentCEvent {self.__class__.__name__}:"
                f"of component={self.component.name} : {self.component.global_id} with priority ={self.priority}\n")



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import copy
import logging

from availsim4core.src.analysis import Analysis
from availsim4core.src.reader.xlsx.monte_carlo_reader import SimulationMonteCarloColumn
from availsim4core.src.reader.xlsx.sensitivity_analysis_reader import SensitivityAnalysisReader
from availsim4core.src.reader.xlsx.system_template_reader import SystemTemplateArchitectureColumn, \
    SystemTemplateFailureModeColumn, SystemTemplateMinimalReplaceableUnitColumn, SystemTemplateInspectionsColumn


class ParameterOfSensitivityAnalysisNotFoundError(Exception):
    pass


class SensitivityAnalysis:
    """
    Manages the sensitivity analysis from the initial simulation and initial system.
    """

    def __init__(self,
                 initial_simulation,
                 initial_system_template):
        self.initial_simulation = initial_simulation
        self.initial_system_template = initial_system_template

    def generate_analysis_list(self,
                               path_sensitivity_analysis):
        """
        From the given path of the sensitivity analysis it reads the file and returns a list of Analysis to perform.
        :param path_sensitivity_analysis path of the configuration file for the sensitivity analysis.
        :return List[Analysis]
        :see Analysis
        """

        system_modifier_combination_list = SensitivityAnalysisReader.generate_system_modifier_combination_list(
            path_sensitivity_analysis)

        analysis_list = []
        for idx, system_modifier_combination in enumerate(system_modifier_combination_list):
            analysis = self._generate_analysis(idx,
                                               system_modifier_combination)
            analysis_list.append(analysis)
        return analysis_list

    def _generate_analysis(self,
                           idx,
                           system_modifier_combination):
        """
        Given a combination {SystemModifierCombination} and based of the initial simulation and system,
        it generates the corresponding analysis.
        :param idx the id of the analysis to generate
        :param system_modifier_combination the SystemModifierCombination to apply on initial simulation and system.
        :returns Corresponding Analysis {Analysis}
        """

        new_simulation = self._apply_combination_modifier_on_simulation(system_modifier_combination)
        new_system_template = self._apply_combination_modifier_on_system(system_modifier_combination)
        return Analysis(idx,
                        new_system_template,
                        new_simulation)

    def _apply_combination_modifier_on_system(self,
                                              system_modifier_combination):
        """
        Given a {SystemModifierCombination} it applies the modification to the copy of the initial system template.
        :param system_modifier_combination the combination with the system modifier to apply on the initial system template.
        """
        modified_initial_system_template = copy.deepcopy(self.initial_system_template)

        for system_modifier in system_modifier_combination.system_modifier_list:

            modification_applied = False

            column_name, name = system_modifier.parameter_name.partition("/")[::2]

            if column_name=="SEED":
                # modification applied in another function
                modification_applied = True

            architecture_entry = modified_initial_system_template.find_architecture_entry(name)
            if architecture_entry is not None:
                if column_name == SystemTemplateArchitectureColumn.COMPONENT_NUMBER:
                    architecture_entry.component_number = int(system_modifier.value)
                    modification_applied = True
                if column_name == SystemTemplateArchitectureColumn.CHILDREN_LOGIC:
                    architecture_entry.children_logic_str = system_modifier.value
                    modification_applied = True

            failure_mode = modified_initial_system_template.find_failure_mode(name)
            if failure_mode is not None:
                if column_name == SystemTemplateFailureModeColumn.FAILURE_PARAMETERS:
                    failure_mode.failure_law.parameters = system_modifier.value
                    modification_applied = True
                if column_name == SystemTemplateFailureModeColumn.REPAIR_PARAMETERS:
                    failure_mode.repair_law.parameters = system_modifier.value
                    modification_applied = True

            mru = modified_initial_system_template.find_mru(name)
            if mru is not None:
                if column_name == SystemTemplateMinimalReplaceableUnitColumn.REPAIR_PARAMETERS:
                    mru.repair_law.parameters = system_modifier.value
                    modification_applied = True
                if column_name == SystemTemplateMinimalReplaceableUnitColumn.REPAIR_SCHEDULE:
                    mru.repair_schedule = system_modifier.value
                    modification_applied = True

            inspection = modified_initial_system_template.find_inspection(name)
            if inspection is not None:
                if column_name == SystemTemplateInspectionsColumn.INSPECTION_DURATION:
                    inspection.duration = float(system_modifier.value)
                    modification_applied = True
                if column_name == SystemTemplateInspectionsColumn.INSPECTION_PERIOD:
                    inspection.periodicity = float(system_modifier.value)
                    modification_applied = True

            if not modification_applied:
                exception_message = (f"System_modifier {system_modifier } could not be applied: wrong parameter name={name}")
                logging.exception(exception_message)
                raise ParameterOfSensitivityAnalysisNotFoundError(exception_message)

        return modified_initial_system_template

    def _apply_combination_modifier_on_simulation(self, system_modifier_combination):
        """
        Given a {SystemModifierCombination} it applies the modification to the copy of the initial simulation.
        :param system_modifier_combination the combination with the system modifier to apply on the initial simulation.
        Note: Because the simulation has only one line entry, only the column name of the system modifier is taken into account.
        """
        modified_initial_simulation = copy.deepcopy(self.initial_simulation)

        for system_modifier in system_modifier_combination.system_modifier_list:
            column_name, *_ = system_modifier.parameter_name.split("/")

            if column_name == SimulationMonteCarloColumn.SEED:
                modified_initial_simulation.seed = int(system_modifier.value)

        return modified_initial_simulation



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import unittest

from availsim4core.src.context.phase.phase import Phase
from availsim4core.src.context.phase.phase_manager import PhaseManager
from availsim4core.src.context.system.probability_law.deterministic_law import DeterministicLaw
from availsim4core.src.timeline.record_phase import RecordPhase


class test_PhaseManager(unittest.TestCase):

    def test__get_first_phase(self):
        phase_first = Phase("phase_test", DeterministicLaw(10), True)
        phase_second = Phase("phase_test", DeterministicLaw(10), False)
        phase_first.set_next_phase(phase_second)
        phase_second.set_next_phase(phase_first)
        phase_set = {phase_first, phase_second}

        phase_manager = PhaseManager(phase_set, set())
        result = phase_manager._get_first_phase()

        expected_result = phase_first
        self.assertEqual(expected_result, result)

    def test__get_first_phase__none_phase(self):
        phase_set = set()

        phase_manager = PhaseManager(phase_set, set())
        result = phase_manager._get_first_phase()

        expected_result = PhaseManager.NO_PHASE

        self.assertEqual(expected_result, result)

    def test_find_phase(self):
        phase_first = Phase("phase_test_first", DeterministicLaw(10), True)
        phase_second = Phase("phase_test_second", DeterministicLaw(10), False)
        phase_first.set_next_phase(phase_second)
        phase_second.set_next_phase(phase_first)
        phase_set = {phase_first, phase_second}

        phase_manager = PhaseManager(phase_set, set())
        result = phase_manager._find_phase("phase_test_first")
        expected_result = phase_first

        self.assertEqual(result, expected_result)

    def test_find_phase_no_phase(self):
        phase_set = set()
        phase_manager = PhaseManager(phase_set, set())

        result = phase_manager._find_phase("not_found_phase")
        expected_result = phase_manager.NO_PHASE

        self.assertEqual(expected_result, result)

    def test_set_next_default_phase__get_first(self):
        phase_first = Phase("phase_test_first", DeterministicLaw(10), True)
        phase_second = Phase("phase_test_second", DeterministicLaw(10), False)
        phase_first.set_next_phase(phase_second)
        phase_second.set_next_phase(phase_first)
        phase_set = {phase_first, phase_second}

        phase_manager = PhaseManager(phase_set, set())

        expected_result_current_phase = phase_first

        self.assertEqual(expected_result_current_phase, phase_manager.current_phase)

    def test_set_next_default_phase__get_second_phase(self):
        phase_first = Phase("phase_test_first", DeterministicLaw(10), True)
        phase_second = Phase("phase_test_second", DeterministicLaw(10), False)
        phase_first.set_next_phase(phase_second)
        phase_second.set_next_phase(phase_first)
        phase_set = {phase_first, phase_second}

        phase_manager = PhaseManager(phase_set, set())

        result = phase_manager.go_to_phase(0,phase_manager.current_phase.next,PhaseManager.DEFAULT_PHASE_DESCRIPTION)

        expected_result = phase_second

        expected_result_phase_record = RecordPhase(phase_second,
                                                   0,
                                                   PhaseManager.DEFAULT_PHASE_DESCRIPTION)

        self.assertEqual(expected_result, phase_manager.current_phase)
        self.assertEqual(expected_result_phase_record, result)

    def test_set_next_default_phase_if_failure(self):
        phase_first = Phase("phase_test_first", DeterministicLaw(10), True)
        phase_second = Phase("phase_test_second", DeterministicLaw(10), False)
        phase_first.set_next_phase(PhaseManager.NO_PHASE)
        phase_second.set_next_phase(PhaseManager.NO_PHASE)
        phase_first.set_next_phase_if_failure(phase_second)
        phase_second.set_next_phase_if_failure(phase_first)
        phase_set = {phase_first, phase_second}

        phase_manager = PhaseManager(phase_set, set())
        result = phase_manager.go_to_phase(0,
                                           phase_second,
                                           PhaseManager.DEFAULT_PHASE_IF_FAILURE_DESCRIPTION)

        expected_result = phase_second

        expected_result_phase_record = RecordPhase(phase_second,
                                                   0,
                                                   PhaseManager.DEFAULT_PHASE_IF_FAILURE_DESCRIPTION)

        self.assertEqual(expected_result, phase_manager.current_phase)
        self.assertEqual(expected_result_phase_record, result)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import os
import random
import shutil
import time
import unittest

import pandas
from memory_profiler import memory_usage

from availsim4core import main


class test_Benchmark(unittest.TestCase):
    """
    Ensemble of tests used to monitored the computation time of the code
    Unfortunately the CI of git.cern.ch is not always using the same hardware to compute the simulations so a large
    margin is used...
    """

    #  set 1st (0: x) value to limit the time in seconds
    #  set 2nd (1: x) value to limit the memory usage in MiB

    expected_result_scenario_0 = {0: 30,  # on CI it is slower ...
                                  1: 160}
    scenario_0_BM11 = \
        ["./availsim4core/test/E2E/input/benchmark/BM_simulation.xlsx",
         "./availsim4core/test/E2E/input/benchmark/BM11_system.xlsx",
         None,
         1,
         expected_result_scenario_0]

    expected_result_scenario_1 = {0: 50,
                                  1: 160}
    scenario_1_BM100 = \
        ["./availsim4core/test/E2E/input/benchmark/BM_simulation_3.xlsx",
         "./availsim4core/test/E2E/input/benchmark/BM100_system.xlsx",
         None,
         1,
         expected_result_scenario_1]

    param_scenario = [
        scenario_0_BM11,
        scenario_1_BM100
    ]

    def test_runner(self):
        # randomize the order of the test execution to detect hidden dependencies
        random.shuffle(self.param_scenario)
        for simulation_file, system_file, sensitivity_file, expected_number_file_result, expected_mean_result in self.param_scenario:
            with self.subTest():
                self._runner_E2E(simulation_file,
                                 system_file,
                                 sensitivity_file,
                                 expected_number_file_result,
                                 expected_mean_result)

    def _runner_E2E(self, simulation_file, system_file, sensitivity_file, expected_number_file_result,
                    expected_mean_result):

        output_folder = "./availsim4core/test/E2E/output/benchmark/"

        # Clean folder
        if os.path.exists(output_folder):
            shutil.rmtree(output_folder)

        # Run the main process
        start_time = time.time()
        main.start(simulation_file, system_file, sensitivity_file, output_folder)
        execution_time = time.time() - start_time
        print(f"bench Execution time = {execution_time}")
        mem_usage = memory_usage(-1, 1, 1)
        memo_element = mem_usage[0]
        profiling = {"execution_time [s]": {0: execution_time},
                     "memory_usage [MiB]": {0: memo_element}}
        profiling_dataframe = pandas.DataFrame(profiling)
        self.assertLessEqual(profiling_dataframe["execution_time [s]"][0], expected_mean_result[0])
        self.assertLessEqual(profiling_dataframe["memory_usage [MiB]"][0], expected_mean_result[1])


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import logging

from availsim4core.src.context.system.failure import Failure, FailureType
from availsim4core.src.context.system.system_utils import SystemUtils


class FailureFactoryError(Exception):
    pass


class FailureFactory:

    @staticmethod
    def build(failure_mode_type_of_failure_str):

        if "DETECTABLE" == failure_mode_type_of_failure_str:
            return Failure(FailureType.DETECTABLE)

        elif "BLIND" == SystemUtils.extract_name_of_function_from_string(failure_mode_type_of_failure_str):
            return Failure(FailureType.BLIND)
        else:
            message = f"Wrong type of failure : {failure_mode_type_of_failure_str} not found "
            logging.exception(message)
            raise FailureFactoryError(message)



# SPDX-License-Identifier: GPL-3.0-only
# (C) Copyright CERN 2021. All rights not expressly granted are reserved. 

import logging
from typing import List

from availsim4core.src.context.context import Context
from availsim4core.src.context.system.children_logic.children_logic import ChildrenLogic
from availsim4core.src.context.system.component_tree.component import Component
from availsim4core.src.context.system.component_tree.status import Status


class ChildrenLogicOoError(Exception):
    pass


class Oo(ChildrenLogic):
    """
    class used to handle an "Out of" logic. "Out Of" means X component(s) out of Y components has(ve) to be in the
    RUNNING state for the parent to be in the RUNNING state
    """

    def __init__(self,
                 minimum_number_of_required_component: int,
                 total_number_of_component=None):
        self.minimum_number_of_required_component = minimum_number_of_required_component
        self.total_number_of_component = total_number_of_component

    def __repr__(self):
        return f"{self.minimum_number_of_required_component}OO{self.total_number_of_component}"

    @staticmethod
    def count_status_occurrence(status: Status, list_of_children: List[Component]) -> int:
        """
        Number of occurrences of a given status from the given list of component.
        :param status: status to evaluate
        :param list_of_children: full list of children
        :return occurrence: int, the occurrence of the given status within the given component list.
        """

        return len([child for child in list_of_children if child.status == status])

    def evaluate(self, list_of_children: List[Component], context: Context) -> Status:

        total_number_of_component = len(list_of_children)
        if self.total_number_of_component is not None:
            if self.total_number_of_component != total_number_of_component:

                list_of_parents_name = []
                for child in list_of_children:
                    for parent in child._parents:
                        list_of_parents_name.append(f"{parent.name}_{parent.local_id}_{parent.global_id}")

                message_exception = f"Total number of children components should be {self.total_number_of_component} " \
                                    f"according to the CHILDREN_LOGIC_INPUT of component(s) {set(list_of_parents_name)} " \
                                    f"but it is equal to {total_number_of_component}"
                logging.exception(message_exception)
                raise ChildrenLogicOoError(message_exception)


        occurrence_by_status = {}

        # search for the presence of different statuses
        for status in [Status.FAILED, Status.BLIND_FAILED, Status.HELD,
                       Status.DEGRADED, Status.BLIND_DEGRADED,
                       Status.INSPECTION, Status.UNDER_REPAIR,
                       Status.RUNNING]:
            occurrence_by_status[status] = self.count_status_occurrence(status, list_of_children)

        considered_as_running = occurrence_by_status[Status.RUNNING] + \
                                occurrence_by_status[Status.DEGRADED] + \
                                occurrence_by_status[Status.BLIND_DEGRADED]

        # RUNNING
        if occurrence_by_status[Status.RUNNING] == total_number_of_component:
            return Status.RUNNING
        else:
            # DEGRADED
            if considered_as_running >= self.minimum_number_of_required_component:

                if occurrence_by_status[Status.BLIND_FAILED] + occurrence_by_status[Status.BLIND_DEGRADED] == 0:
                    return Status.DEGRADED
                elif occurrence_by_status[Status.BLIND_FAILED] + occurrence_by_status[Status.BLIND_DEGRADED] > 0:
                    return Status.BLIND_DEGRADED
                else:
                    raise ChildrenLogicOoError(
                        "Unexpected negative occurrence of Statuses in Oo children logic - Degraded")

            elif considered_as_running < self.minimum_number_of_required_component:

                if occurrence_by_status[Status.INSPECTION] > 0:
                    return Status.INSPECTION
                elif occurrence_by_status[Status.INSPECTION] == 0 and occurrence_by_status[Status.UNDER_REPAIR] > 0:
                    return Status.UNDER_REPAIR
                elif occurrence_by_status[Status.INSPECTION] + occurrence_by_status[Status.UNDER_REPAIR] == 0 \
                        and occurrence_by_status[Status.BLIND_FAILED] + occurrence_by_status[Status.BLIND_DEGRADED] > 0:
                    return Status.BLIND_FAILED
                elif considered_as_running + occurrence_by_status[
                    Status.HELD] >= self.minimum_number_of_required_component:
                    return Status.HELD
                elif occurrence_by_status[Status.INSPECTION] + occurrence_by_status[Status.UNDER_REPAIR] \
                        + occurrence_by_status[Status.BLIND_FAILED] + occurrence_by_status[Status.BLIND_DEGRADED] == 0:
                    # Default Case
                    return Status.FAILED
                else:
                    raise ChildrenLogicOoError(
                        f"Unexpected occurrence of Statuses in Oo children logic over the components {list_of_children}"
                        )
            else:
                message_exception = f"Unexpected occurrence of Statuses in Oo children logic"
                logging.exception(message_exception)
                raise ChildrenLogicOoError(message_exception)
